from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tabulous.widgets import TableBase, TableViewerBase, Table, SpreadSheet

__void = object()


def get_table(viewer: TableViewerBase, default=__void) -> TableBase:
    """Safely get the current table of the given viewer."""
    if table := viewer.current_table:
        return table
    if default is not __void:
        return default
    raise ValueError("No table found in the viewer.")


def get_mutable_table(viewer: TableViewerBase, default=__void) -> Table | SpreadSheet:
    """Safely get the current mutable-type table of the given viewer."""
    try:
        table = get_table(viewer)
    except ValueError as e:
        if default is not __void:
            return default
        raise e

    if table.table_type in ("Table", "SpreadSheet"):
        return table

    if default is not __void:
        return default
    from tabulous.exceptions import TableImmutableError

    raise TableImmutableError(f"Table {table.name!r} is immutable.")


def get_spreadsheet(viewer: TableViewerBase, default=__void) -> SpreadSheet:
    """Safely get the current mutable-type table of the given viewer."""
    try:
        table = get_table(viewer)
    except ValueError as e:
        if default is not __void:
            return default
        raise e
    if table.table_type == "SpreadSheet":
        return table
    if default is not __void:
        return default
    raise TypeError(f"Table {table.name!r} is not a spreadsheet.")


def get_selected_row(viewer: TableViewerBase) -> int:
    table = get_table(viewer)
    selected = table.index.selected
    if len(selected) == 0:
        raise ValueError("No index selected")
    if len(selected) > 1:
        raise ValueError("Multiple indices selected")

    return selected[0].start


def get_selected_column(viewer: TableViewerBase) -> int:
    table = get_table(viewer)
    selected = table.columns.selected
    if len(selected) == 0:
        raise ValueError("No columns selected")
    if len(selected) > 1:
        raise ValueError("Multiple ranges are selected")
    first = selected[0]
    if first.stop != first.start + 1:
        raise ValueError("Multiple columns are selected")

    return first.start


def get_table_and_column_name(viewer: TableViewerBase) -> tuple[TableBase, str]:
    table = get_table(viewer)
    selected = table.columns.selected
    if len(selected) == 0:
        raise ValueError("No columns selected")
    if len(selected) > 1:
        raise ValueError("Multiple ranges are selected")
    first = selected[0]
    if first.stop != first.start + 1:
        raise ValueError("Multiple columns are selected")

    return table, table.columns[first.start]


def get_selected_columns(
    viewer: TableViewerBase, assert_exists: bool = True
) -> list[int]:
    table = get_table(viewer)
    selected = table.columns.selected
    if assert_exists and len(selected) == 0:
        raise ValueError("No columns selected")

    out: list[int] = []
    for sl in selected:
        out.extend(range(sl.start, sl.stop))
    return out


def get_a_selection(table: TableBase) -> tuple[slice, slice]:
    """Get a selection of the given table."""
    sels = table.selections
    if len(sels) != 1:
        raise ValueError("Multiple selections are selected")
    return sels[0]


def get_clipboard_text() -> str:
    from qtpy import QtWidgets as QtW

    s = QtW.QApplication.clipboard().text()
    if s:
        return s
    raise ValueError("No text found in clipboard.")



from __future__ import annotations

from typing import TYPE_CHECKING
from pathlib import Path
from dataclasses import dataclass
import weakref

if TYPE_CHECKING:
    from tabulous.widgets._table import TableBase

_void = object()


@dataclass(frozen=True)
class Source:
    """Class that describes source info of a table."""

    path: Path | None = None
    parent: weakref.ReferenceType[TableBase] | None = None

    @classmethod
    def from_table(self, table: TableBase) -> Source:
        return Source(parent=weakref.ref(table))

    def get_parent(self) -> TableBase | None:
        if self.parent is not None:
            return self.parent()
        return None

    def replace(self, path=_void, parent=_void) -> Source:
        if path is _void:
            path = self.path
        if parent is _void:
            parent = self.parent
        return Source(path=path, parent=parent)



from __future__ import annotations

from abc import abstractmethod
import ast
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, overload
import numpy as np
from enum import Enum
from functools import reduce
from tabulous.types import ProxyType, _IntArray, _IntOrBoolArray
from tabulous.exceptions import TableNotOrderedError

if TYPE_CHECKING:
    import pandas as pd
    from typing_extensions import Self


class ProxyTypes(Enum):
    none = "none"
    unknown = "unknown"
    filter = "filter"
    sort = "sort"

    def __eq__(self, other: ProxyTypes | str) -> bool:
        if isinstance(other, str):
            return self.value == other
        return super().__eq__(other)


class SortFilterProxy:
    """A custom sort/filter proxy for pandas dataframes."""

    def __init__(self, obj: ProxyType | None = None):
        if isinstance(obj, SortFilterProxy):
            obj = obj._obj
        if isinstance(obj, Composable) and obj.is_identity():
            obj = None
        self._obj: ProxyType | None = obj
        if self._obj is None:
            self._proxy_type = ProxyTypes.none
            self._is_ordered = True
        else:
            self._proxy_type = ProxyTypes.unknown
            self._is_ordered = False
        self._last_indexer = None

    def __repr__(self) -> str:
        cname = type(self).__name__
        return f"{cname}<proxy_type={self.proxy_type}, obj={self._obj!r}>"

    @property
    def proxy_type(self) -> ProxyTypes:
        """The proxy type."""
        return self._proxy_type

    @property
    def is_ordered(self) -> bool:
        return self._is_ordered

    @property
    def last_indexer(self) -> _IntOrBoolArray | None:
        return self._last_indexer

    def apply(
        self,
        df: pd.DataFrame,
        ref: pd.DataFrame | Callable[[], pd.DataFrame] | None = None,
    ) -> pd.DataFrame:
        """
        Apply the proxy rule to the dataframe.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to be sliced.
        ref : pd.DataFrame, optional
            The reference dataframe to be used to determine the slice, if proxy is
            callable. If not given, ``df`` will be used. If a callable is given,
            it will be called to supply the reference dataframe.
        """
        sl = self._obj
        if sl is None:
            return df
        # get indexer
        if callable(sl):
            if ref is None:
                ref_input = df
            elif callable(ref):
                ref_input = ref()
            sl_filt = sl(ref_input)
        else:
            sl_filt = sl

        self._last_indexer = sl_filt

        if self._array_is_bool(sl_filt):
            df_filt = df[sl_filt]
        else:
            df_filt = df.iloc[sl_filt]
        return df_filt

    # fmt: off
    @overload
    def get_source_index(self, r: int) -> int: ...
    @overload
    def get_source_index(self, r: slice) -> slice | _IntArray: ...
    @overload
    def get_source_index(self, r: list[int]) -> _IntArray: ...
    # fmt: on

    def get_source_index(self, r):
        """Get the source index of the row in the dataframe."""
        sl = self._obj
        if sl is None:
            if isinstance(r, list):
                r = np.array(r)
            r0 = r
        else:
            if callable(sl):
                if self._last_indexer is not None:
                    sl = self._last_indexer
                else:
                    raise RuntimeError("Call apply first!")

            if self._array_is_bool(sl):
                sl = np.where(sl)[0]
            r0 = sl[r]
            if isinstance(r0, np.integer):
                r0 = int(r0)
        return r0

    def get_source_slice(self, r: slice, force_single_row: bool = False) -> slice:
        """Get the source row slice in the dataframe."""
        if self.proxy_type is ProxyTypes.none:
            return r
        start, stop = r.start, r.stop
        if start is None:
            start = 0
        if stop is None:
            if self._last_indexer is None:
                raise ValueError("Cannot determine stop index")
            stop = self._last_indexer.size
        if force_single_row and start != stop - 1:
            raise TableNotOrderedError("Only single-row selection is allowed.")
        start, stop = self.get_source_index([start, stop - 1])
        return slice(start, stop + 1)

    def map_slice(self, r: slice) -> slice:
        if self.proxy_type is ProxyTypes.none:
            return r
        start, stop = r.start, r.stop
        if start is None:
            start = 0
        if stop is None:
            if self._last_indexer is None:
                raise ValueError("Cannot determine stop index")
            stop = self._last_indexer.size
        # NOTE: mapping is not defined for all sources. For example, filter slice
        # [True, False, True] cannot map 1 because it does not exist in the
        # filtered data.
        if self._array_is_bool(self._last_indexer):
            _cumsum = np.cumsum(self._last_indexer)
            return slice(max(_cumsum[start] - 1, 0), max(_cumsum[stop - 1], 0))
        elif start == stop - 1:
            indices = np.where(self._last_indexer == start)[0]
            if len(indices) > 0:
                idx = indices[0]
                return slice(idx, idx + 1)
            raise ValueError(f"Cannot map slice {r} to source")
        else:
            raise TableNotOrderedError("Cannot map slice if proxy is not ordered.")

    def as_indexer(self, df: pd.DataFrame | None) -> _IntOrBoolArray | slice:
        sl = self._obj
        if sl is None:
            return slice(None)
        if callable(sl):
            if self._last_indexer is not None:
                sl_filt = self._last_indexer
            else:
                if df is None:
                    raise ValueError("Cannot determine indexer")
                sl_filt = sl(df)
        else:
            sl_filt = sl
        return sl_filt

    def _array_is_bool(self, sl: _IntOrBoolArray) -> bool:
        if sl.dtype.kind == "b":
            self._proxy_type = ProxyTypes.filter
            self._is_ordered = True
        elif sl.dtype.kind in "ui":
            self._proxy_type = ProxyTypes.sort
            self._is_ordered = False
        else:
            raise TypeError(f"Invalid filter type: {sl.dtype}")
        return self._is_ordered


class FilterType(Enum):
    """Filter type enumeration."""

    none = "none"
    eq = "eq"
    ne = "ne"
    gt = "gt"
    ge = "ge"
    lt = "lt"
    le = "le"
    between = "between"
    isin = "isin"
    startswith = "startswith"
    endswith = "endswith"
    contains = "contains"
    matches = "matches"

    @property
    def repr(self) -> str:
        return _REPR_MAP[self]

    @property
    def requires_number(self) -> bool:
        cls = type(self)
        return self in {cls.eq, cls.ne, cls.gt, cls.ge, cls.lt, cls.le}

    @property
    def requires_text(self) -> bool:
        cls = type(self)
        return self in {
            cls.between,
            cls.startswith,
            cls.endswith,
            cls.contains,
            cls.matches,
        }

    @property
    def requires_list(self) -> bool:
        return self is FilterType.isin

    @classmethod
    def from_ast(self, obj: ast.AST):
        if isinstance(obj, ast.Eq):
            return FilterType.eq
        elif isinstance(obj, ast.NotEq):
            return FilterType.ne
        elif isinstance(obj, ast.Gt):
            return FilterType.gt
        elif isinstance(obj, ast.GtE):
            return FilterType.ge
        elif isinstance(obj, ast.Lt):
            return FilterType.lt
        elif isinstance(obj, ast.LtE):
            return FilterType.le
        else:
            raise ValueError(f"Invalid AST object: {obj}")

    def __str__(self) -> str:
        return self.repr


def _is_between(x: pd.Series, a: str) -> pd.Series:
    a = a.strip()
    left = a[0]
    right = a[-1]
    values = [float(s.strip()) for s in a[1:-1].split(",")]
    if left == "[" and right == "]":
        inclusive = "both"
    elif left == "[" and right == ")":
        inclusive = "left"
    elif left == "(" and right == "]":
        inclusive = "right"
    elif left == "(" and right == ")":
        inclusive = "neither"
    else:
        raise ValueError(f"Invalid range: {a}")
    if len(values) != 2:
        raise ValueError(f"Invalid range: {a}")
    return x.between(*values, inclusive=inclusive)


_FUNCTION_MAP: dict[FilterType, Callable[[pd.Series, Any], pd.Series]] = {
    FilterType.none: lambda x, a: np.ones(len(x), dtype=bool),
    FilterType.eq: lambda x, a: x == a,
    FilterType.ne: lambda x, a: x != a,
    FilterType.gt: lambda x, a: x > a,
    FilterType.ge: lambda x, a: x >= a,
    FilterType.lt: lambda x, a: x < a,
    FilterType.le: lambda x, a: x <= a,
    FilterType.between: _is_between,
    FilterType.isin: lambda x, a: x.isin(a),
    FilterType.startswith: lambda x, a: x.str.startswith(a),
    FilterType.endswith: lambda x, a: x.str.endswith(a),
    FilterType.contains: lambda x, a: x.str.contains(a),
    FilterType.matches: lambda x, a: x.str.contains(a, regex=True),
}

_REPR_MAP: dict[FilterType, str] = {
    FilterType.none: "Select ...",
    FilterType.eq: "=",
    FilterType.ne: "≠",
    FilterType.gt: ">",
    FilterType.ge: "≥",
    FilterType.lt: "<",
    FilterType.le: "≤",
    FilterType.isin: "is in",
    FilterType.startswith: "starts with",
    FilterType.between: "between",
    FilterType.endswith: "ends with",
    FilterType.contains: "contains",
    FilterType.matches: ".*",
}


class FilterInfo(NamedTuple):
    type: FilterType
    arg: Any


class Composable:
    @abstractmethod
    def __call__(self, df: pd.DataFrame) -> np.ndarray:
        """Apply the mapping to the dataframe."""

    @abstractmethod
    def copy(self) -> Self:
        """Copy the instance."""

    @abstractmethod
    def indices(self) -> set[int]:
        """Get the indices of the header targets."""

    @abstractmethod
    def compose(self, column: int) -> Self:
        """Compose with an additional mapping at the column."""

    @abstractmethod
    def decompose(self, column: int) -> Self:
        """Decompose the mapping at column."""

    @abstractmethod
    def is_identity(self) -> bool:
        """True if this instance is the identity mapping."""


class ComposableFilter(Composable):
    def __init__(self, d: dict[int, FilterInfo] | None = None):
        if d is not None:
            assert isinstance(d, dict)
            self._dict: dict[int, FilterInfo] = d
        else:
            self._dict = {}
        self.__name__ = "filter"
        self.__annotations__ = {"df": "pd.DataFrame", "return": np.ndarray}

    @classmethod
    def from_ast(cls, obj: ast.Compare, columns: pd.Index) -> Self:
        if len(obj.ops) != 1:
            raise ValueError("AST has more than one operator")
        op = obj.ops[0]
        left = obj.left
        right = obj.comparators[0]
        if isinstance(left, ast.Name) and isinstance(right, ast.Constant):
            colname, other = left, right
        elif isinstance(left, ast.Constant) and isinstance(right, ast.Name):
            colname, other = right, left
        else:
            raise ValueError("Must compare a variable and a constant")
        _info = FilterInfo(FilterType.from_ast(op), other.value)
        index = columns.get_loc(colname.id)
        return cls({index: _info})

    def __call__(self, df: pd.DataFrame) -> np.ndarray:
        series: list[pd.Series] = []
        if len(self._dict) == 0:
            return np.ones(len(df), dtype=bool)
        for index, (type, arg) in self._dict.items():
            fn = _FUNCTION_MAP[type]
            series.append(np.asarray(fn(df.iloc[:, index], arg)))
        return reduce(lambda x, y: x & y, series)

    def copy(self) -> ComposableFilter:
        """Copy the filter object."""
        return self.__class__(self._dict.copy())

    def indices(self) -> set[int]:
        return set(self._dict.keys())

    def compose(self, column: int, info: FilterInfo) -> ComposableFilter:
        """Compose with an additional column filter."""
        if info.type is FilterType.none:
            return self
        new = self.copy()
        new._dict[column] = info
        return new

    def decompose(self, column: int) -> ComposableFilter:
        """Decompose the filter at column."""
        new = self.copy()
        new._dict.pop(column, None)
        return new

    def is_identity(self) -> bool:
        """True if the filter is the identity filter."""
        return len(self._dict) == 0


class ComposableSorter(Composable):
    def __init__(self, columns: set[int] | None = None, ascending: bool = True):
        if columns is None:
            self._columns: set[int] = set()
        else:
            assert isinstance(columns, set)
            self._columns = columns
        self._ascending = ascending

    def __call__(self, df: pd.DataFrame) -> pd.Series:
        by: list[str] = [df.columns[i] for i in self._columns]
        if len(by) == 1:
            out = np.asarray(df[by[0]].argsort())
            if not self._ascending:
                out = out[::-1]
        else:
            df_sub = df[by]
            nr = len(df_sub)
            df_sub.index = range(nr)
            df_sub = df_sub.sort_values(by=by, ascending=self._ascending)
            out = np.asarray(df_sub.index)
        return out

    def copy(self) -> ComposableSorter:
        return self.__class__(self._columns.copy(), self._ascending)

    def indices(self) -> set[int]:
        return self._columns.copy()

    def compose(self, column: int):
        """Compose the sorter object."""
        new = self.copy()
        new._columns.add(column)
        return new

    def decompose(self, column: int):
        """Decompose the filter object."""
        new = self.copy()
        new._columns.remove(column)
        return new

    def switch(self) -> ComposableSorter:
        """New sorter with the reverse order."""
        return self.__class__(self._columns, not self._ascending)

    def is_identity(self) -> bool:
        """True if the sorter is the identity sorter."""
        return len(self._columns) == 0



from __future__ import annotations

from typing import (
    Any,
    Callable,
    Generic,
    Hashable,
    TypeVar,
    TYPE_CHECKING,
    overload,
    Literal,
)

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase, TableBase
    from typing_extensions import Self

_T = TypeVar("_T")


class _Joinable:
    def __init__(self, name: str = ""):
        self._name = name
        self._instances: dict[Hashable, Self] = {}

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if (out := self._instances.get(instance, None)) is None:
            out = self._instances[instance] = self.__class__(self._name)
        return out

    def __set_name__(self, owner, name):
        self._name = name

    def _join(self, other: Self):
        raise NotImplementedError()


class _Registerable(_Joinable):
    def __init__(self, name: str = ""):
        super().__init__(name)
        self._registered: list[Any] = []

    def _register(self, loc, func=None):
        def wrapper(f):
            self._registered.append((loc, f))
            return f

        return wrapper(func) if func is not None else wrapper

    def _join(self, other: Self):
        self._registered.extend(other._registered)


class _Updatable(_Joinable):
    def __init__(self, name: str = ""):
        super().__init__(name)
        self._dict = {}

    def __setitem__(self, key: str, value: Any):
        self._dict[key] = value

    def update(self, *args, **kwargs):
        self._dict.update(*args, **kwargs)

    def add(self, obj: _T) -> _T:
        """A decorator to add an callable object to the namespace."""
        if callable(obj) or isinstance(obj, type):
            name = obj.__name__
            self[name] = obj
        else:
            raise TypeError(f"Expected to be used as a decorator, got {type(obj)}")
        return obj

    def _join(self, other: Self):
        self._dict.update(other._dict)


class ContextRegisterable(_Registerable, Generic[_T]):
    """A mock for the `register` method."""

    # fmt: off
    @overload
    def register(self, loc: str, func: Literal[None] = None) -> Callable[[Callable[[_T, Any], None]], Callable[[_T, Any], None]]: ...  # noqa: E501
    @overload
    def register(self, loc: str, func: Callable[[_T, Any], None]) -> Callable[[_T, Any], None]: ...  # noqa: E501
    # fmt: on

    def register(self, loc, func=None):
        return self._register(loc, func)


class KeyMapMock(_Registerable):
    """A mock for the `keymap` attribute of a table viewer instance."""

    # fmt: off
    @overload
    def register(self, key: str, func: Literal[None] = None) -> Callable[[Callable[[TableViewerBase, Any], None]], Callable[[TableViewerBase, Any], None]]: ...  # noqa: E501
    @overload
    def register(self, key: str, func: Callable[[TableViewerBase, Any], None]) -> Callable[[TableViewerBase, Any], None]: ...  # noqa: E501
    # fmt: on

    def register(self, key, func=None):
        return self._register(key, func)


class CellNamespaceMock(_Updatable):
    """A mock for the `cell_namespace` attribute of a table viewer instance."""


class ConsoleMock(_Updatable):
    """A mock for the `console` attribute of a table viewer instance."""


class CommandPaletteMock(_Registerable):
    # fmt: off
    @overload
    def register(self, command: Callable[[TableViewerBase], Any], title: str, desc: str | None, key: str | None): ...  # noqa: E501
    @overload
    def register(self, command: Literal[None], title: str, desc: str | None, key: str | None): ...  # noqa: E501
    # fmt: on

    def register(self, command, title="User defined", desc=None, key=None):
        return self._register((command, title, desc, key))


class Initializer:
    _fields = ()

    def __hash__(self):
        return id(self)

    def join(self, other: Initializer):
        """Join initializers together."""
        for name in self._fields:
            self_field: _Registerable = getattr(self, name)
            other_field: _Registerable = getattr(other, name)
            self_field._join(other_field)
        return self


class ViewerInitializer(Initializer):
    tables: ContextRegisterable[TableViewerBase] = ContextRegisterable()
    keymap: ContextRegisterable[TableViewerBase] = KeyMapMock()
    console: ConsoleMock[TableViewerBase] = ConsoleMock()
    cell_namespace: CellNamespaceMock[TableViewerBase] = CellNamespaceMock()
    command_palette: CommandPaletteMock[TableViewerBase] = CommandPaletteMock()

    _fields = ("tables", "keymap", "console", "cell_namespace", "command_palette")

    def initialize_viewer(self, viewer: TableViewerBase):
        for args in self.tables._registered:
            viewer.tables.register(*args)
        for args in self.keymap._registered:
            # NOTE: The QtKeyMap object is currently a class variable. When the second
            # viewer is launched, old keybindings are still registered. To avoid this,
            # we just allow overwriting the keymap.
            viewer.keymap.register(*args, overwrite=True)
        viewer.cell_namespace.update_safely(self.cell_namespace._dict)
        viewer.console.update(self.console._dict)
        for args in self.command_palette._registered:
            viewer.command_palette.register(*args)


class TableInitializer(Initializer):
    cell: ContextRegisterable[TableBase] = ContextRegisterable()
    index: ContextRegisterable[TableBase] = ContextRegisterable()
    columns: ContextRegisterable[TableBase] = ContextRegisterable()
    keymap: ContextRegisterable[TableBase] = KeyMapMock()

    _fields = ("cell", "index", "columns", "keymap")

    def initialize_table(self, table: TableBase):
        for args in self.cell._registered:
            table.cell.register(*args)
        for args in self.index._registered:
            table.index.register(*args)
        for args in self.columns._registered:
            table.columns.register(*args)
        for args in self.keymap._registered:
            table.keymap.register(*args, overwrite=True)


_VIEWER_INITIALIZER = ViewerInitializer()
_TABLE_INITIALIZER = TableInitializer()


def get_initializers() -> tuple[ViewerInitializer, TableInitializer]:
    """Get viewer and table initializers."""
    return _VIEWER_INITIALIZER, _TABLE_INITIALIZER



from __future__ import annotations

from typing import TYPE_CHECKING
from . import _utils

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase


def set_popup_mode(viewer: TableViewerBase):
    """Popup table."""
    table = _utils.get_table(viewer)
    table.view_mode = "popup"


def set_dual_v_mode(viewer: TableViewerBase):
    """Vertical dual-view"""
    table = _utils.get_table(viewer)
    table.view_mode = "vertical"


def set_dual_h_mode(viewer: TableViewerBase):
    """Horizontal dual-view"""
    table = _utils.get_table(viewer)
    table.view_mode = "horizontal"


def reset_view_mode(viewer: TableViewerBase):
    """Reset view mode"""
    table = _utils.get_table(viewer)
    table.view_mode = "normal"



from __future__ import annotations
from typing import TYPE_CHECKING
from qt_command_palette import get_storage

if TYPE_CHECKING:
    from tabulous._qt._mainwindow import _QtMainWidgetBase


def register_storage_vars():
    storage = get_storage("tabulous")

    @storage.mark_getter
    def viewer(self: _QtMainWidgetBase):
        return self._table_viewer

    @storage.mark_getter
    def table(self: _QtMainWidgetBase):
        return self._table_viewer.current_table



from __future__ import annotations

from types import FunctionType
from typing import MutableMapping, Any, TypeVar, Mapping

_T = TypeVar("_T", FunctionType, type)


class Namespace(MutableMapping[str, Any]):
    """Namespace used in cell edit"""

    def __init__(self):
        import numpy, pandas

        self._ns: dict[str, Any] = {"np": numpy, "pd": pandas}
        self._static = set(self._ns.keys())

    def __len__(self) -> int:
        return len(self._ns)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._ns!r})"

    def __iter__(self):
        return iter(self._ns)

    def _check_static(self, key: str):
        if key in self._static:
            raise ValueError(f"Cannot update {key!r}")

    def __getitem__(self, key: str) -> Any:
        return self._ns[key]

    def __setitem__(self, key: str, val: Any) -> None:
        self._check_static(key)
        self._ns[key] = val

    def __delitem__(self, key: str) -> None:
        self._check_static(key)
        del self._ns[key]

    def update(self, ns: Mapping[str, Any] = {}, /, **kwargs) -> None:
        ns = dict(**ns, **kwargs)
        if collision := set(ns.keys()) & self._static:
            raise ValueError(f"Cannot update {collision!r}")
        return self._ns.update(ns)

    def update_safely(self, ns: Mapping[str, Any] = {}, /, **kwargs) -> None:
        ns = dict(**ns, **kwargs)
        for k in self._static:
            ns.pop(k, None)
        return self._ns.update(ns)

    def add(self, obj: _T) -> _T:
        """A decorator to add an callable object to the namespace."""
        if callable(obj) or isinstance(obj, type):
            name = obj.__name__
            self[name] = obj
        else:
            raise TypeError(f"Expected to be used as a decorator, got {type(obj)}")
        return obj



from __future__ import annotations
from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt, Signal

if TYPE_CHECKING:
    from ._enhanced_table import _QTableViewEnhanced

# Wrapper widgets that can be used to wrap a QTableView


class QTableDualView(QtW.QSplitter):
    """Dual view of the same table."""

    def __init__(
        self,
        table: _QTableViewEnhanced,
        orientation=Qt.Orientation.Horizontal,
    ):
        super().__init__(orientation)
        self.setChildrenCollapsible(False)

        second = table.copy(link=True)

        self.addWidget(table)
        self.addWidget(second)
        self.setSizes([500, 500])

        self._table = table
        self._second = second

    def deleteLater(self) -> None:
        self._second._selection_model.moving.disconnect(self._second._on_moving)
        self._second._selection_model.moved.disconnect(self._second._on_moved)
        return super().deleteLater()


class QTablePopupView(QtW.QWidget):
    """Popup view widget for a table."""

    def __init__(self, table: _QTableViewEnhanced):
        super().__init__()
        self.setLayout(QtW.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(table)

        self._table = table
        self._second = table.copy(link=True)

        popup = QPopupWidget(table, self._second)
        self.popup = popup
        self.popup.closed.connect(lambda: self._second.setParent(None))

    def exec(self):
        self.popup.show()
        self.popup._widget.setFocus()

    def deleteLater(self) -> None:
        self._second._selection_model.moving.disconnect(self._second._on_moving)
        self._second._selection_model.moved.disconnect(self._second._on_moved)
        self.popup.close()
        return super().deleteLater()


class QPopupWidget(QtW.QWidget):
    closed = Signal()

    def __init__(
        self,
        widget: _QTableViewEnhanced = None,
        parent: _QTableViewEnhanced = None,
    ):
        super().__init__(parent, Qt.WindowType.Popup)
        self._widget = widget
        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(4, 4, 4, 4)

        self._title = QtW.QLabel("")
        self._title.setFont(QtGui.QFont("Arial", 14, QtGui.QFont.Weight.Bold))
        self._title.setAlignment(Qt.AlignmentFlag.AlignCenter)
        _layout.addWidget(self._title)
        _layout.addWidget(widget)
        sizegrip = QtW.QSizeGrip(self)
        _layout.addWidget(sizegrip, False, Qt.AlignmentFlag.AlignRight)

        self.setLayout(_layout)
        _screen_rect = QtGui.QGuiApplication.primaryScreen().geometry()
        _screen_center = _screen_rect.center()
        self.resize(int(_screen_rect.width() * 0.8), int(_screen_rect.height() * 0.8))
        self.move(_screen_center - self.rect().center())
        self._drag_start: QtCore.QPoint | None = None

    def setTitle(self, text: str):
        """Set the title of the popup."""
        return self._title.setText(text)

    def closeEvent(self, a0) -> None:
        self.closed.emit()
        return super().closeEvent(a0)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        self._drag_start = event.pos()
        return super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        self._drag_start = None
        return super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        if self._drag_start is not None:
            self.move(self.mapToParent(event.pos() - self._drag_start))

        return super().mouseMoveEvent(event)



from functools import partial
from typing import Union, Iterable
from typing_extensions import Annotated, Literal
from sklearn import cluster, svm, linear_model, decomposition
from sklearn.mixture import GaussianMixture


class ModelRegistry(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._hidden = {}

    def register(self, name: str):
        """Register a model class to the registry."""

        def _wrapper(val):
            self[name] = val
            return val

        return _wrapper

    def hide_keys(self, keys: Iterable[str]) -> None:
        """Hide given keys from the registry."""
        self._hidden = {key: self.pop(key) for key in keys}
        return None

    def show_keys(self, keys: Union[Iterable[str], None] = None) -> None:
        """Show given keys from the registry."""
        if keys is None:
            keys = list(self._hidden.keys())
        self.update({key: self._hidden.pop(key) for key in keys})
        return None


MODELS = ModelRegistry()


def _normalize_random_state(state) -> Union[int, None]:
    if state:
        return int(state)
    else:
        return None


# Model factories
_AffinityType = Literal["euclidean", "l1", "l2", "manhattan", "cosine", "precomputed"]
_BoolOrAuto = Literal["auto", "true", "false"]


@MODELS.register("k-means")
def kmeans(
    n_clusters: int = 8,
    init: Literal["k-means++", "random"] = "k-means++",
    n_init: Annotated[int, {"min": 1, "max": 100}] = 10,
    max_iter: Annotated[int, {"min": 1, "max": 10000}] = 300,
    tol: str = "1e-4",
    random_state: str = "",
):
    random_state = _normalize_random_state(random_state)
    tol = float(tol)
    return cluster.KMeans(
        n_clusters=n_clusters,
        init=init,
        n_init=n_init,
        max_iter=max_iter,
        tol=tol,
        random_state=random_state,
    )


@MODELS.register("DBSCAN")
def dbscan(
    eps: Annotated[float, {"min": 0.0, "max": 1000.0}] = 0.5,
    min_samples: Annotated[int, {"min": 1, "max": 100}] = 5,
    metric: str = "euclidean",
    algorithm: Literal["auto", "ball_tree", "kd_tree", "brute"] = "auto",
    leaf_size: Annotated[int, {"min": 1, "max": 100}] = 30,
    p: Annotated[int, {"min": 1, "max": 100}] = 2,
    n_jobs: Annotated[int, {"min": 1, "max": 100}] = None,
):
    return cluster.DBSCAN(
        eps=eps,
        min_samples=min_samples,
        metric=metric,
        algorithm=algorithm,
        leaf_size=leaf_size,
        p=p,
        n_jobs=n_jobs,
    )


@MODELS.register("Agglomerative")
def agglomerative(
    n_clusters: Annotated[int, {"min": 1, "max": 100}] = 2,
    affinity: _AffinityType = "euclidean",
    compute_full_tree: _BoolOrAuto = "auto",
    linkage: Literal["ward", "complete", "average", "single"] = "ward",
):
    return cluster.AgglomerativeClustering(
        n_clusters=n_clusters,
        affinity=affinity,
        compute_full_tree=compute_full_tree,
        linkage=linkage,
    )


@MODELS.register("OPTICS")
def optics(
    min_samples: Annotated[int, {"min": 1, "max": 100}] = 5,
    max_eps: str = "inf",
    metric: str = "minkowski",
    algorithm: Literal["auto", "ball_tree", "kd_tree", "brute"] = "auto",
    leaf_size: Annotated[int, {"min": 1, "max": 100}] = 30,
    p: Annotated[int, {"min": 1, "max": 100}] = 2,
    n_jobs: Annotated[int, {"min": 1, "max": 100}] = None,
):
    max_eps = float(max_eps)
    return cluster.OPTICS(
        min_samples=min_samples,
        max_eps=max_eps,
        metric=metric,
        algorithm=algorithm,
        leaf_size=leaf_size,
        p=p,
        n_jobs=n_jobs,
    )


@MODELS.register("Spectral")
def spectral(
    n_clusters: Annotated[int, {"min": 1, "max": 100}] = 8,
    eigen_solver: Literal["arpack", "lobpcg", "amg"] = "arpack",
    random_state: str = "",
    n_init: Annotated[int, {"min": 1, "max": 100}] = 10,
    gamma: Annotated[float, {"min": 0.0, "max": 100.0}] = 1.0,
    affinity: Literal["nearest_neighbors", "rbf"] = "rbf",
    n_neighbors: Annotated[int, {"min": 1, "max": 100}] = 10,
    eigen_tol: Annotated[float, {"min": 0.0, "max": 100.0}] = 0.0,
    assign_labels: Literal["kmeans", "discretize"] = "kmeans",
    degree: float = 3,
    coef0: float = 1,
):
    random_state = _normalize_random_state(random_state)
    return cluster.SpectralClustering(
        n_clusters=n_clusters,
        eigen_solver=eigen_solver,
        random_state=random_state,
        n_init=n_init,
        gamma=gamma,
        affinity=affinity,
        n_neighbors=n_neighbors,
        eigen_tol=eigen_tol,
        assign_labels=assign_labels,
        degree=degree,
        coef0=coef0,
    )


@MODELS.register("Affinity propagation")
def affinity_propagation(
    damping: Annotated[float, {"min": 0.5, "max": 1.0, "step": 0.01}] = 0.5,
    max_iter: Annotated[int, {"min": 1, "max": 10000}] = 200,
    convergence_iter: Annotated[int, {"min": 1, "max": 10000}] = 15,
    affinity: _AffinityType = "euclidean",
):
    return cluster.AffinityPropagation(
        damping=damping,
        max_iter=max_iter,
        convergence_iter=convergence_iter,
        affinity=affinity,
    )


@MODELS.register("Birch")
def birch(
    n_clusters: Annotated[int, {"min": 1, "max": 100}] = 3,
    threshold: Annotated[float, {"min": 0.0, "max": 100.0}] = 0.5,
    branching_factor: Annotated[int, {"min": 1, "max": 100}] = 50,
    compute_labels: _BoolOrAuto = "auto",
):
    return cluster.Birch(
        n_clusters=n_clusters,
        threshold=threshold,
        branching_factor=branching_factor,
        compute_labels=compute_labels,
    )


@MODELS.register("Mean shift")
def mean_shift(
    bandwidth: Annotated[float, {"min": 0.0, "max": 100.0}] = None,
    bin_seeding: bool = False,
    min_bin_freq: Annotated[int, {"min": 1, "max": 100}] = 1,
    cluster_all: bool = True,
):
    return cluster.MeanShift(
        bandwidth=bandwidth,
        bin_seeding=bin_seeding,
        min_bin_freq=min_bin_freq,
        cluster_all=cluster_all,
    )


@MODELS.register("Gaussian mixture")
def gaussian_mixture(
    n_components: Annotated[int, {"min": 1, "max": 100}] = 1,
    covariance_type: Literal["full", "tied", "diag", "spherical"] = "full",
    tol: str = "1e-3",
    reg_covar: Annotated[float, {"min": 0.0, "max": 100.0}] = 1e-6,
    max_iter: Annotated[int, {"min": 1, "max": 1000}] = 100,
    n_init: Annotated[int, {"min": 1, "max": 100}] = 1,
    init_params: Literal["kmeans", "random"] = "kmeans",
    random_state: str = "",
):
    random_state = _normalize_random_state(random_state)
    tol = float(tol)
    return GaussianMixture(
        n_components=n_components,
        covariance_type=covariance_type,
        tol=tol,
        reg_covar=reg_covar,
        max_iter=max_iter,
        n_init=n_init,
        init_params=init_params,
        random_state=random_state,
    )


_SvmKernels = Literal["linear", "poly", "rbf", "sigmoid", "precomputed"]


@MODELS.register("SVC")
def svc(
    C: Annotated[float, {"min": 0.0, "max": 100.0}] = 1.0,
    kernel: _SvmKernels = "rbf",
    degree: Annotated[int, {"min": 1, "max": 100}] = 3,
    # gamma: _GammaType = "scale",
    coef0: Annotated[float, {"min": 0.0, "max": 100.0}] = 0.0,
    shrinking: bool = True,
    probability: bool = False,
    tol: str = "1e-3",
    cache_size: Annotated[int, {"min": 1, "max": 1000}] = 200,
    random_state: str = "",
    regression: bool = False,
):
    random_state = _normalize_random_state(random_state)
    if regression:
        svm_cls = svm.SVR
    else:
        svm_cls = svm.SVC
    return svm_cls(
        C=C,
        kernel=kernel,
        degree=degree,
        # gamma=gamma,
        coef0=coef0,
        shrinking=shrinking,
        probability=probability,
        tol=float(tol),
        cache_size=cache_size,
        random_state=random_state,
    )


@MODELS.register("Nu-SVC")
def nusvc(
    nu: Annotated[float, {"min": 0.0, "max": 1.0}] = 0.5,
    kernel: _SvmKernels = "rbf",
    degree: Annotated[int, {"min": 1, "max": 100}] = 3,
    # gamma: _GammaType = "scale",
    coef0: Annotated[float, {"min": 0.0, "max": 100.0}] = 0.0,
    shrinking: bool = True,
    tol: str = "1e-3",
    cache_size: Annotated[int, {"min": 1, "max": 100}] = 200,
    random_state: str = "",
    regression: bool = False,
):
    random_state = _normalize_random_state(random_state)
    if regression:
        svm_cls = svm.NuSVR
    else:
        svm_cls = svm.NuSVC
    return svm_cls(
        nu=nu,
        kernel=kernel,
        degree=degree,
        # gamma=gamma,
        coef0=coef0,
        shrinking=shrinking,
        tol=float(tol),
        cache_size=cache_size,
        random_state=random_state,
    )


@MODELS.register("Linear-SVC")
def linear_svc(
    penalty: Literal["l1", "l2"] = "l2",
    dual: bool = True,
    tol: str = "1e-4",
    C: Annotated[float, {"min": 0.0, "max": 100.0}] = 1.0,
    multi_class: Literal["ovr", "crammer_singer"] = "ovr",
    fit_intercept: bool = True,
    intercept_scaling: Annotated[float, {"min": 0.0, "max": 100.0}] = 1,
    random_state: str = "",
    max_iter: Annotated[int, {"min": 1, "max": 10000}] = 1000,
    regression: bool = False,
):
    random_state = _normalize_random_state(random_state)
    if regression:
        if penalty == "l1":
            loss = "epsilon_insensitive"
        else:
            loss = "squared_epsilon_insensitive"
        svm_cls = partial(svm.LinearSVR, loss=loss)
    else:
        svm_cls = partial(svm.LinearSVC, penalty=penalty)
    return svm_cls(
        penalty=penalty,
        dual=dual,
        tol=float(tol),
        C=C,
        multi_class=multi_class,
        fit_intercept=fit_intercept,
        intercept_scaling=intercept_scaling,
        random_state=random_state,
        max_iter=max_iter,
    )


@MODELS.register("Ridge")
def ridge(
    alpha: Annotated[float, {"min": 0.0, "max": 100.0}] = 1.0,
    fit_intercept: bool = True,
    normalize: bool = False,
    max_iter: Annotated[int, {"min": 0, "max": 10000}] = 0,
    tol: str = "1e-3",
    solver: Literal[
        "auto", "svd", "cholesky", "lsqr", "sparse_cg", "sag", "saga"
    ] = "auto",
    random_state: str = "",
):
    random_state = _normalize_random_state(random_state)
    if max_iter == 0:
        max_iter = None
    return linear_model.Ridge(
        alpha=alpha,
        fit_intercept=fit_intercept,
        normalize=normalize,
        max_iter=max_iter,
        tol=float(tol),
        solver=solver,
        random_state=random_state,
    )


@MODELS.register("Lasso")
def lasso(
    alpha: Annotated[float, {"min": 0.0, "max": 100.0}] = 1.0,
    fit_intercept: bool = True,
    normalize: bool = False,
    max_iter: Annotated[int, {"min": 0, "max": 10000}] = 1000,
    tol: str = "1e-4",
    warm_start: bool = False,
    positive: bool = False,
    random_state: str = "",
    selection: Literal["cyclic", "random"] = "cyclic",
):
    random_state = _normalize_random_state(random_state)
    return linear_model.Lasso(
        alpha=alpha,
        fit_intercept=fit_intercept,
        normalize=normalize,
        max_iter=max_iter,
        tol=float(tol),
        warm_start=warm_start,
        positive=positive,
        random_state=random_state,
        selection=selection,
    )


@MODELS.register("ElasticNet")
def elastic_net(
    alpha: Annotated[float, {"min": 0.0, "max": 100.0}] = 1.0,
    l1_ratio: Annotated[float, {"min": 0.0, "max": 1.0}] = 0.5,
    fit_intercept: bool = True,
    normalize: bool = False,
    max_iter: Annotated[int, {"min": 0, "max": 10000}] = 1000,
    tol: str = "1e-4",
    warm_start: bool = False,
    positive: bool = False,
    random_state: str = "",
    selection: Literal["cyclic", "random"] = "cyclic",
):
    random_state = _normalize_random_state(random_state)
    return linear_model.ElasticNet(
        alpha=alpha,
        l1_ratio=l1_ratio,
        fit_intercept=fit_intercept,
        normalize=normalize,
        max_iter=max_iter,
        tol=float(tol),
        warm_start=warm_start,
        positive=positive,
        random_state=random_state,
        selection=selection,
    )


@MODELS.register("Multi-task Lasso")
def multi_task_lasso(
    alpha: Annotated[float, {"min": 0.0, "max": 100.0}] = 1.0,
    fit_intercept: bool = True,
    normalize: bool = False,
    max_iter: Annotated[int, {"min": 0, "max": 10000}] = 1000,
    tol: str = "1e-4",
    warm_start: bool = False,
    random_state: str = "",
    selection: Literal["cyclic", "random"] = "cyclic",
):
    random_state = _normalize_random_state(random_state)
    return linear_model.MultiTaskLasso(
        alpha=alpha,
        fit_intercept=fit_intercept,
        normalize=normalize,
        max_iter=max_iter,
        tol=float(tol),
        warm_start=warm_start,
        random_state=random_state,
        selection=selection,
    )


@MODELS.register("PCA")
def pca(
    n_components: Annotated[int, {"min": 1, "max": 1000}] = 2,
    whiten: bool = False,
    svd_solver: Literal["auto", "full", "arpack", "randomized"] = "auto",
    tol: str = "0.0",
    iterated_power: Annotated[int, {"min": 1, "max": 1000}] = 5,
    random_state: str = "",
):
    random_state = _normalize_random_state(random_state)
    return decomposition.PCA(
        n_components=n_components,
        whiten=whiten,
        svd_solver=svd_solver,
        tol=float(tol),
        iterated_power=iterated_power,
        random_state=random_state,
    )


@MODELS.register("FactICA")
def fact_ica(
    n_components: Annotated[int, {"min": 1, "max": 1000}] = 2,
    algorithm: Literal["parallel", "deflation"] = "parallel",
    whiten: bool = False,
    fun: Literal["logcosh", "exp", "cube"] = "logcosh",
    max_iter: Annotated[int, {"min": 1, "max": 1000}] = 200,
    tol: str = "1e-4",
):
    return decomposition.FastICA(
        n_components=n_components,
        algorithm=algorithm,
        whiten=whiten,
        fun=fun,
        max_iter=max_iter,
        tol=float(tol),
    )


BASIC = {"k-means", "Gaussian mixture", "Ridge", "PCA"}
ADVANCED = set(MODELS.keys()).difference(BASIC)



from __future__ import annotations
from typing import TYPE_CHECKING

from qtpy import QtWidgets as QtW

if TYPE_CHECKING:
    from ._overlay import QOverlayFrame

# TODO:
class QObjectHolder(QtW.QWidget):
    def __init__(self, parent: QtW.QWidget = None):
        super().__init__(parent)
        _layout = QtW.QVBoxLayout()
        self.setLayout(_layout)

        from tabulous._qt import QTableLayer
        import pandas as pd

        qtable = QTableLayer(
            data=pd.DataFrame([[""]] * 12, columns=["Object"], dtype=object)
        )
        qtable.setTextFormatter("Object", lambda x: repr(x))

        _layout.addWidget(qtable)

    def overlayWidget(self) -> QOverlayFrame:
        return self.parent()

    def parentTable(self):
        return self.overlayWidget().parentTable()

    def parentTableView(self):
        return self.parentTable()._qtable_view

    def close_widget(self):
        qtable = self.parentTableView()
        qtable._selection_model.moved.disconnect(self._on_selection_changed)
        self.hide()
        del qtable._focused_widget
        self.deleteLater()
        qtable.setFocus()
        return None

    def _on_selection_changed(self):
        ...



from __future__ import annotations
from functools import partial

from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt, Signal
from tabulous.style import Style, GLOBAL_STYLES
from tabulous._qt._mainwindow import QMainWindow
from tabulous._utils import get_config, update_config


class QThemePanel(QtW.QWidget):
    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)
        self._labels: dict[str, QThemeSelectionLabel] = {}
        _layout = QtW.QGridLayout()
        row = 0
        col = 0
        ncols = 6

        for name, style in GLOBAL_STYLES.items():
            label = QThemeSelectionLabel(style)
            label.setToolTip(name)
            self._labels[name] = label
            label.clicked.connect(partial(self.setTheme, name=name))
            _layout.addWidget(label, row, col)
            col += 1
            if col >= ncols:
                row += 1
                col = 0

        self.setLayout(_layout)
        self.setMinimumSize(200, 100)
        name = get_config().window.theme
        self._update_check_state(name)

    def setTheme(self, name: str):
        cfg = get_config()
        cfg.window.theme = name
        update_config(cfg)
        QMainWindow.reload_config()
        self._update_check_state(name)

    def _update_check_state(self, name: str):
        for key, wdt in self._labels.items():
            wdt.setChecked(key == name)


class QThemeSelectionLabel(QtW.QLabel):
    clicked = Signal()

    def __init__(self, style: Style) -> None:
        super().__init__()
        self._style_theme = style
        self.setFixedSize(30, 30)
        self.setFont(QtGui.QFont("Arial", 16))
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self._checked = False

    def setChecked(self, checked: bool):
        self._checked = checked
        self.update()

    def mouseReleaseEvent(self, ev: QtGui.QMouseEvent) -> None:
        self.clicked.emit()
        return super().mouseReleaseEvent(ev)

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(30, 30)

    def paintEvent(self, a0: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self)
        geo = self.rect()

        grad = QtGui.QLinearGradient(geo.topLeft(), geo.bottomRight())
        grad.setColorAt(0, QtGui.QColor(self._style_theme.background0))
        grad.setColorAt(1, QtGui.QColor(self._style_theme.background1))
        path = QtGui.QPainterPath(geo.topLeft())
        path.lineTo(geo.topRight())
        path.lineTo(geo.bottomLeft())
        painter.fillPath(path, grad)

        grad = QtGui.QLinearGradient(geo.topLeft(), geo.bottomRight())
        grad.setColorAt(0, QtGui.QColor(self._style_theme.highlight0))
        grad.setColorAt(1, QtGui.QColor(self._style_theme.highlight1))
        path = QtGui.QPainterPath(geo.topRight())
        path.lineTo(geo.bottomLeft())
        path.lineTo(geo.bottomRight())
        painter.fillPath(path, grad)

        if self._checked:
            painter.setPen(QtGui.QPen(QtGui.QColor(255, 20, 20), 3))
            painter.drawRect(geo)

        painter.setPen(QtGui.QPen(QtGui.QColor(self._style_theme.foreground), 3))
        painter.drawText(geo, Qt.AlignmentFlag.AlignCenter, "A")
        return None



from __future__ import annotations

from enum import Enum
from scipy import stats


class Distributions(Enum):
    """Choices of distributions"""

    norm = "Normal"
    expon = "Exponential"
    uniform = "Uniform"
    gamma = "Gamma"
    beta = "Beta"
    chi2 = "χ^2"
    t = "T"
    f = "F"
    lognorm = "log-Normal"
    poisson = "Poisson"
    binom = "Binomial"
    boltzmann = "Boltzmann"

    def __str__(self) -> str:
        return self.value

    @property
    def dist(self) -> stats.rv_continuous | stats.rv_discrete:
        """Get the distribution"""
        return getattr(stats, self.name)

    @property
    def params(self) -> tuple[str, ...]:
        """Get the parameters"""
        return _PARAM_NAMES[self]

    @property
    def latex(self) -> str:
        """Get the LaTeX formula"""
        return _LATEX_FORMULAS[self]


_PARAM_NAMES = {
    Distributions.norm: ("loc", "scale"),
    Distributions.expon: ("loc", "scale"),
    Distributions.uniform: ("loc", "scale"),
    Distributions.gamma: ("a", "loc", "scale"),
    Distributions.beta: ("a", "b", "loc", "scale"),
    Distributions.chi2: ("df", "loc", "scale"),
    Distributions.t: ("df", "loc", "scale"),
    Distributions.f: ("dfn", "dfd", "loc", "scale"),
    Distributions.lognorm: ("s", "loc", "scale"),
    Distributions.poisson: ("mu",),
    Distributions.binom: ("n", "p"),
    Distributions.boltzmann: ("T", "loc", "scale"),
}

# fmt: off
_LATEX_FORMULAS = {
    Distributions.norm: r"$f(x) = \frac{1}{\sqrt{2\pi\sigma^2}}\exp\left(-\frac{(x-\mu)^2}{2\sigma^2}\right)$",  # noqa: E501
    Distributions.expon: r"$f(x) = \lambda e^{-\lambda x}$",  # noqa: E501
    Distributions.uniform: r"$f(x) = \frac{1}{b-a}$",  # noqa: E501
    Distributions.gamma: r"$f(x) = \frac{x^{k-1}e^{-x/\theta}}{\theta^k\Gamma(k)}$",  # noqa: E501
    Distributions.beta: r"$f(x) = \frac{x^{a-1}(1-x)^{b-1}}{B(a,b)}$",  # noqa: E501
    Distributions.chi2: r"$f(x) = \frac{1}{2^{\nu/2}\Gamma(\nu/2)}x^{\nu/2-1}e^{-x/2}$",  # noqa: E501
    Distributions.t: r"$f(x) = \frac{\Gamma((\nu+1)/2)}{\sqrt{\pi\nu}\Gamma(\nu/2)}\left(1+\frac{x^2}{\nu}\right)^{-(\nu+1)/2}$",  # noqa: E501
    Distributions.f: r"$f(x) = \frac{(\nu_1 x)^{\nu_1}\nu_2^{\nu_2}}{\nu_1 x+\nu_2}\frac{\Gamma(\nu_1+\nu_2)}{\Gamma(\nu_1)\Gamma(\nu_2)}\frac{1}{x}$",  # noqa: E501
    Distributions.lognorm: r"$f(x) = \frac{1}{x\sigma\sqrt{2\pi}}\exp\left(-\frac{(\ln x-\mu)^2}{2\sigma^2}\right)$",  # noqa: E501
    Distributions.poisson: r"$f(x) = \frac{\lambda^k e^{-\lambda}}{k!}$",  # noqa: E501
    Distributions.binom: r"$f(x) = \binom{n}{k}p^k(1-p)^{n-k}$",  # noqa: E501
    Distributions.boltzmann: r"$f(x) = \frac{1}{Z}\exp\left(-\frac{x-\mu}{\theta}\right)\left(\frac{x-\mu}{\theta}\right)^{k-1}$",  # noqa: E501
}
# fmt: on



from __future__ import annotations

from qtpy import QtWidgets as QtW
from magicgui.backends._qtpy.widgets import QBaseWidget
from magicgui.widgets import Widget
from tabulous.widgets import TableViewerWidget
from tabulous.types import TabPosition


class MagicTableViewer(Widget, TableViewerWidget):
    """
    A magicgui widget of table viewer.

    This class is a subclass of ``magicgui.widget.Widget`` so that it can be used
    in a compatible way with magicgui and napari.

    Parameters
    ----------
    tab_position: TabPosition or str
        Type of list-like widget to use.
    """

    def __init__(
        self,
        *,
        tab_position: TabPosition | str = TabPosition.top,
        name: str = "",
        label: str = None,
        tooltip: str | None = None,
        visible: bool | None = None,
        enabled: bool = True,
        show: bool = False,
    ):
        super().__init__(
            widget_type=QBaseWidget,
            backend_kwargs={"qwidg": QtW.QWidget},
            name=name,
            label=label,
            tooltip=tooltip,
            visible=visible,
            enabled=enabled,
        )
        TableViewerWidget.__init__(self, tab_position=tab_position, show=False)
        mgui_native: QtW.QWidget = self._widget._mgui_get_native_widget()
        mgui_native.setLayout(QtW.QVBoxLayout())
        mgui_native.layout().addWidget(self._qwidget)
        mgui_native.setContentsMargins(0, 0, 0, 0)
        if show:
            self.show(run=False)

    @property
    def native(self):
        try:
            return TableViewerWidget.native.fget(self)
        except AttributeError:
            return Widget.native.fget(self)



from __future__ import annotations

from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt

if TYPE_CHECKING:
    from tabulous._qt._mainwindow import _QtMainWidgetBase


class QIntOrEmptyValidator(QtGui.QIntValidator):
    def validate(self, input: str, pos: int) -> tuple[QtGui.QValidator.State, str, int]:
        if input == "":
            return QtGui.QValidator.State.Acceptable, input, pos
        return super().validate(input, pos)


class QJumpWidget(QtW.QDialog):
    """Implements the jump-to behavior."""

    def __init__(self, parent: _QtMainWidgetBase | None = None) -> None:
        super().__init__(parent, Qt.WindowType.SubWindow)
        _line_edits = QtW.QWidget()
        self._msg = QtW.QLabel("")

        self._row = QtW.QLineEdit()
        self._row.setToolTip("Row index")
        self._col = QtW.QLineEdit()
        self._col.setToolTip("Column index")
        idx = parent._table_viewer.current_table.current_index
        self._row.setText(str(idx[0]))
        self._col.setText(str(idx[1]))
        self._row.setValidator(QIntOrEmptyValidator())
        self._col.setValidator(QIntOrEmptyValidator())

        self._row.textChanged.connect(self._on_text_changed)
        self._col.textChanged.connect(self._on_text_changed)
        _hbox = QtW.QHBoxLayout()
        _hbox.addWidget(self._row)
        _hbox.addWidget(QtW.QLabel(", "))
        _hbox.addWidget(self._col)
        _hbox.setContentsMargins(0, 0, 0, 0)
        _line_edits.setLayout(_hbox)

        _vbox = QtW.QVBoxLayout()
        _vbox.addWidget(QtW.QLabel("Jump to:"))
        _vbox.addWidget(_line_edits)
        _vbox.addWidget(self._msg)
        self.setLayout(_vbox)

        self._row.editingFinished.connect(self._on_row_finished)
        self._col.editingFinished.connect(self._on_col_finished)
        self._last_valid_row, self._last_valid_col = idx

        # move to the center of the parent
        self.resize(240, 100)
        self.move(parent.rect().center() - self.rect().center())

    def _delete(self):
        self.hide()
        self.deleteLater()

    def _on_row_finished(self):
        self._col.setFocus()
        self._col.selectAll()
        if not self._col.hasFocus():
            self._delete()

    def _on_col_finished(self):
        self.parentWidget().setCellFocus()
        if not self._row.hasFocus():
            self._delete()

    def _on_text_changed(self):
        rtext = self._row.text()
        ctext = self._col.text()
        if rtext == "":
            r = self._last_valid_row
        else:
            self._last_valid_row = r = int(rtext)
        if ctext == "":
            c = self._last_valid_col
        else:
            self._last_valid_col = c = int(ctext)

        table = self.parentWidget()._table_viewer.current_table
        if r < table.table_shape[0] and c < table.table_shape[1]:
            table.move_iloc(r, c)
            self._msg.setText("")
        else:
            self._msg.setText(f"Index out of bound: ({r}, {c})")

    def parentWidget(self) -> _QtMainWidgetBase:
        return super().parentWidget()

    def show(self):
        super().show()
        self._row.selectAll()
        self._row.setFocus()



from __future__ import annotations
from typing import TYPE_CHECKING, Any, cast
import qtpy
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import QEvent, Signal, Qt
from qt_command_palette import get_palette

from ._namespace import Namespace

from tabulous._qt._table import QSpreadSheet
from tabulous._qt._table_stack import QTabbedTableStack
from tabulous._qt._history import QtFileHistoryManager
from tabulous._keymap import QtKeyMap
from tabulous.types import TabPosition
from tabulous._utils import load_cell_namespace
from tabulous.style import Style

if TYPE_CHECKING:
    from tabulous._qt._toolbar import QTableStackToolBar
    from tabulous._qt._mainwindow._titlebar import QMainWindowTitleBar
    from tabulous._qt._console import QtConsole
    from tabulous.widgets import TableViewer


class _EventFilter(QtCore.QObject):
    styleChanged = Signal()

    def eventFilter(self, obj: QtCore.QObject, event: QEvent):
        _type = event.type()
        if _type == QEvent.Type.StyleChange:
            self.styleChanged.emit()
        return False


class _QtMainWidgetBase(QtW.QWidget):
    _table_viewer: TableViewer
    _tablestack: QTabbedTableStack
    _toolbar: QTableStackToolBar
    _menubar: QMainWindowTitleBar
    _keymap: QtKeyMap
    _namespace: Namespace

    _hist_mgr = QtFileHistoryManager()
    _is_mainwindow = False

    def __init__(
        self,
        tab_position: TabPosition | str = TabPosition.top,
    ):
        super().__init__()
        self.setObjectName(f"tabulous.{type(self).__name__}")
        tab_position = TabPosition(tab_position)
        self._tablestack = QTabbedTableStack(tab_position=tab_position.name)
        self._toolbar = None
        self._menubar = None
        self._style_theme = "none"
        self.setCentralWidget(self._tablestack)

        # NOTE: Event filter must be stored as an attribute, otherwise it will be
        # garbage collected.
        self._event_filter = _EventFilter()
        self.installEventFilter(self._event_filter)
        self._white_background = True
        self._event_filter.styleChanged.connect(self.updateWidgetStyle)
        self._console_widget: QtConsole | None = None
        self._keymap_widget = None

        # Queued namespaces for console
        self._queued_ns: dict[str, Any] = {}

        # update with user namespace
        self._namespace = Namespace()
        self._namespace.update_safely(load_cell_namespace())

        # install command palette
        self._command_palette = get_palette("tabulous")
        try:
            self._command_palette.install(self)
        except RuntimeError:
            # During test, main window might have been destroyed
            pass
        else:
            qcommand_palette = self._command_palette.get_widget(self)
            qcommand_palette.hidden.connect(self._on_hidden)
            qcommand_palette.setFont(QtGui.QFont("Arial", 10))

    def _on_hidden(self):
        try:
            self.setCellFocus()
        except AttributeError:
            pass

    def showCommandPalette(self):
        self._command_palette.show_widget(self)
        return None

    def backgroundColor(self):
        return self.palette().color(self.backgroundRole())

    def applyTheme(self, theme: str):
        """Apply predefined theme to the main window."""
        if self._style_theme == theme:
            return
        self._style_theme = theme
        self.setStyleSheet(Style.from_global(self._style_theme).format_file())
        # update console theme
        if console := self._console_widget:
            console.update_theme(theme)

    def showPreferenceDialog(self):
        from tabulous._qt._preference import QPreferenceDialog

        dlg = QPreferenceDialog(self)
        dlg.exec()

    def updateWidgetStyle(self):
        bg = self.backgroundColor()
        if self._style_theme.startswith("dark"):
            _white = False
        elif self._style_theme.startswith("light"):
            _white = True
        else:
            _white = bg.red() + bg.green() + bg.blue() > 128 * 3
        self._white_background = _white
        if _white:
            color = "#1E1E1E"
        else:
            color = "#CCCCCC"
        # update toolbar tool button colors
        if self._toolbar is not None:
            self._toolbar.setToolButtonColor(color)

        # custom title bar
        if self._menubar is not None:
            self._menubar.setIconColor(color)

        # update header button colors
        if self._tablestack.isEmpty():
            return
        for i in range(self._tablestack.count()):
            table = self._tablestack.tableAtIndex(i)
            if isinstance(table, QSpreadSheet):
                table = cast(QSpreadSheet, table)
                table.model()._out_of_bound_color_cache = None
            for wdt in table._header_widgets().values():
                if hasattr(wdt, "updateColor"):
                    wdt.updateColor(color)

    def screenshot(self):
        """Create an array of pixel data of the current view."""
        import numpy as np

        img = self.grab().toImage()
        bits = img.constBits()
        h, w, c = img.height(), img.width(), 4
        if qtpy.API_NAME.startswith("PySide"):
            arr = np.array(bits).reshape(h, w, c)
        else:
            bits.setsize(h * w * c)
            arr: np.ndarray = np.frombuffer(bits, np.uint8)
            arr = arr.reshape(h, w, c)

        return arr[:, :, [2, 1, 0, 3]]

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        """Activate keymap object."""
        if self._keymap.press_key(a0):
            return None
        return super().keyPressEvent(a0)

    def showKeyMap(self) -> None:
        """Show keymap viewer widget."""
        if self._keymap_widget is None:
            wdt = self._keymap.to_widget()
            wdt.setParent(self, wdt.windowFlags())
            self._keymap_widget = wdt
        self._keymap_widget.show()
        self._keymap_widget.showNormal()
        return None

    def setCellFocus(self) -> None:
        """Set focus to the current table."""
        table = self._table_viewer.current_table
        if table is None or table.table_shape[0] * table.table_shape[1] == 0:
            return None
        sels = table.selections
        if len(sels) == 0:
            table.selections = [(slice(0, 1), slice(0, 1))]

        # set focus
        idx = self._tablestack.currentIndex()
        if len(self._tablestack.tiledIndices(idx)) > 1:
            idx_group = self._tablestack._tab_index_to_group_index(idx)
            self._tablestack.widget(idx).setFocusedIndex(idx_group)
        else:
            table._qwidget._qtable_view.setFocus()
        return None

    def setCentralWidget(self, wdt: QTabbedTableStack):
        """Set the splitter widget."""
        raise NotImplementedError()

    def toolBarVisible(self) -> bool:
        """Visibility of toolbar"""
        raise NotImplementedError()

    def setToolBarVisible(self, visible: bool):
        """Set visibility of toolbar"""
        raise NotImplementedError()

    def toggleToolBarVisibility(self):
        """Toggle visibility of toolbar"""
        return self.setToolBarVisible(not self.toolBarVisible())

    def addDefaultToolBar(self):
        """Add default toolbar widget regardless of self is a main window or not."""
        raise NotImplementedError()

    def consoleVisible(self) -> bool:
        """True if embeded console is visible."""
        raise NotImplementedError()

    def setConsoleVisible(self, visible: bool) -> None:
        """Set visibility of embeded console widget."""
        raise NotImplementedError()

    def toggleConsoleVisibility(self) -> None:
        """Toggle visibility of embeded console widget."""
        return self.setConsoleVisible(not self.consoleVisible())

    def statusBar(self) -> QtW.QStatusBar:
        raise NotImplementedError()

    def toggleWindowState(self):
        if self.windowState() == Qt.WindowState.WindowMaximized:
            self.setWindowState(Qt.WindowState.WindowNoState)
        else:
            self.setWindowState(Qt.WindowState.WindowMaximized)
        if self._menubar is not None:
            self._menubar._corner_buttons.setMiddleIcon(self.windowState())

    def setTableSelection(self, sl: tuple[slice, slice]):
        """Set selection slices to the table selection editor"""
        if table := self._table_viewer.current_table:
            sel = list(table.selections)
            self._toolbar.blockSignals(True)
            sel[-1] = sl
            table.selections = sel
            self._toolbar.blockSignals(False)

    # subclasses will override this method. Just for signature compatibility.
    def close(self, ask: bool | None = False) -> bool:
        """Close the window."""
        return super().close()



from ._table import MagicTableViewer
from ._register import (
    find_table_viewer_ancestor,
    find_current_table,
)
from ._selection import SelectionWidget
from ._dialog import dialog_factory, dialog_factory_mpl
from ._color_edit import ColorEdit
from ._toggle_switch import ToggleSwitch, ToggleSwitches, ToggleSwitchSelect
from ._timedelta import TimeDeltaEdit

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from matplotlib.axes import Axes
else:
    from typing import Any

    Axes = Any

__all__ = [
    "MagicTableViewer",
    "find_table_viewer_ancestor",
    "find_current_table",
    "SelectionWidget",
    "dialog_factory",
    "dialog_factory_mpl",
    "ColorEdit",
    "ToggleSwitch",
    "ToggleSwitches",
    "ToggleSwitchSelect",
    "Axes",
    "TimeDeltaEdit",
]



from typing import Callable, Iterable

from qtpy import QtWidgets as QtW, QtGui
from superqt import QEnumComboBox

from magicgui import magic_factory
from ._distribution import Distributions
from ._latex import QLatexLabel
from tabulous._selection_op import SelectionOperator
from tabulous._magicgui import find_table_viewer_ancestor


class QScipyStatsWidget(QtW.QWidget):
    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        _layout = QtW.QVBoxLayout()
        self._dist = QEnumComboBox(enum_class=Distributions)
        self._dist.setCurrentEnum(Distributions.norm)
        self._latex_label = QLatexLabel(Distributions.norm.latex)
        if parent is not None:
            self._latex_label.setTextColor(
                parent.palette().color(QtGui.QPalette.ColorRole.Text)
            )
        self._params = QStatsParameterWidget(Distributions.norm.params)
        _layout.addWidget(self._dist)
        _layout.addWidget(self._latex_label)
        _area = QtW.QScrollArea()
        _area.setContentsMargins(0, 0, 0, 0)
        _area.setWidget(self._params)
        _area.setFixedHeight(68)
        _layout.addWidget(_area)
        _layout.addWidget(_button("Fit", self.fit))
        _layout.addWidget(_button("Random sampling", self.sample))
        _layout.addWidget(_button("Calculate PDF", self.pdf))
        _layout.addWidget(_button("Calculate CDF", self.cdf))

        self.setLayout(_layout)
        self._dist.currentEnumChanged.connect(self._dist_changed)

    def _dist_changed(self, dist: Distributions):
        self._latex_label.setLatex(dist.latex)
        self._params.set_labels(dist.params)

    def get_instance(self, frozen: bool = True):
        dist: Distributions = self._dist.currentEnum()
        if frozen:
            return dist.dist(*self._params.get_params())
        return dist.dist

    def get_viewer(self):
        return find_table_viewer_ancestor(self)

    def fit(self):
        viewer = self.get_viewer()
        mgui = get_selection()
        mgui.native.setParent(self, mgui.native.windowFlags())
        mgui.show()

        @mgui.called.connect
        def _on_called(sel: SelectionOperator):
            dist = self.get_instance(frozen=False)
            df = sel.operate(viewer.current_table.data_shown)
            params_fit = dist.fit(df.values)
            self._params.set_params(params_fit)
            mgui.close()

    def sample(self):
        """Generate random samples from the distribution."""
        viewer = self.get_viewer()
        mgui = get_selection(selection={"allow_out_of_bounds": True})
        mgui.native.setParent(self, mgui.native.windowFlags())
        mgui.show()

        @mgui.called.connect
        def _on_called(sel: SelectionOperator):
            dist = self.get_instance()
            rsel, csel = sel.as_iloc(None)
            if rsel.stop is None or csel.stop is None:
                raise ValueError(
                    "Unknown shape of selection. You must specify the stop values "
                    "of the slices."
                )
            if rsel.step is not None or csel.step is not None:
                raise ValueError("Step values are not allowed.")
            rsize = rsel.stop - (rsel.start or 0)
            csize = csel.stop - (csel.start or 0)
            out = dist.rvs(rsize * csize).reshape((rsize, csize))
            viewer.current_table.cell[rsel, csel] = out
            mgui.close()

    def cdf(self):
        """Calculate the Cumulative Distribution Function (CDF)."""
        viewer = self.get_viewer()
        df = viewer.current_table.data_shown
        mgui = get_xy_selections(y={"allow_out_of_bounds": True})
        mgui.native.setParent(self, mgui.native.windowFlags())
        mgui.show()

        @mgui.called.connect
        def _on_called(xy: tuple[SelectionOperator, SelectionOperator]):
            xsel, ysel = xy
            dist = self.get_instance()
            ds = xsel.operate(df)
            out = dist.cdf(ds.values.ravel())
            viewer.current_table.cell[ysel.as_iloc(df)] = out
            mgui.close()

    def pdf(self):
        """Calculate the Probability Density Function (PDF)."""
        viewer = self.get_viewer()
        df = viewer.current_table.data_shown
        mgui = get_xy_selections(y={"allow_out_of_bounds": True})
        mgui.native.setParent(self, mgui.native.windowFlags())
        mgui.show()

        @mgui.called.connect
        def _on_called(xy: tuple[SelectionOperator, SelectionOperator]):
            xsel, ysel = xy
            dist = self.get_instance()
            ds = xsel.operate(df)
            out = dist.pdf(ds.values.ravel())
            viewer.current_table.cell[ysel.as_iloc(df)] = out
            mgui.close()


class QStatsParameterWidget(QtW.QWidget):
    def __init__(
        self, labels: Iterable[str], parent: QtW.QWidget | None = None
    ) -> None:
        super().__init__(parent)
        _layout = QtW.QFormLayout()
        self._widgets: list[QtW.QLineEdit] = []
        self.setLayout(_layout)
        self.set_labels(labels)

    def get_params(self) -> list[float]:
        return [float(wdt.text()) for wdt in self._widgets]

    def set_params(self, param: Iterable[float]) -> None:
        for wdt, p in zip(self._widgets, param):
            wdt.setText(str(p))

    def set_labels(self, labels: Iterable[str]) -> None:
        layout: QtW.QFormLayout = self.layout()
        while layout.rowCount() > 0:
            layout.removeRow(0)
        self._widgets: list[QtW.QLineEdit] = []
        for label in labels:
            wdt = QtW.QLineEdit()
            self._widgets.append(wdt)
            layout.addRow(label, wdt)


def _button(text: str, slot: Callable) -> QtW.QPushButton:
    btn = QtW.QPushButton(text)
    btn.clicked.connect(slot)
    btn.setToolTip(slot.__doc__)
    return btn


@magic_factory
def get_selection(selection: SelectionOperator):
    return selection


@magic_factory
def get_xy_selections(
    x: SelectionOperator,
    y: SelectionOperator,
):
    return x, y



from __future__ import annotations
from typing import Iterable
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt, Signal

from magicgui import __version__ as MAGICGUI_VERSION

if int(MAGICGUI_VERSION.split(".")[1]) >= 7:
    from magicgui.widgets.bases import ValueWidget
else:
    from magicgui.widgets._bases import ValueWidget
from magicgui.backends._qtpy.widgets import QBaseValueWidget
from magicgui.application import use_app

from tabulous.color import normalize_color, rgba_to_str


# modified from napari/_qt/widgets/qt_color_swatch.py
class QColorSwatch(QtW.QFrame):
    colorChanged = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self._qcolor: QtGui.QColor = QtGui.QColor(255, 255, 255, 255)
        self.colorChanged.connect(self._update_swatch_style)
        self.setMinimumWidth(40)

    @property
    def rgba(self) -> tuple[int, int, int, int]:
        """Get RBGA tuple from QColor."""
        return self._qcolor.getRgb()

    def heightForWidth(self, w: int) -> int:
        return int(w * 0.667)

    def _update_swatch_style(self, _=None) -> None:
        rgba = f'rgba({",".join(str(x) for x in self.rgba)})'
        return self.setStyleSheet("QColorSwatch {background-color: " + rgba + ";}")

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        """Show QColorPopup picker when the user clicks on the swatch."""
        if event.button() == Qt.MouseButton.LeftButton:
            dlg = QtW.QColorDialog(self._qcolor, self)
            dlg.setOptions(QtW.QColorDialog.ColorDialogOption.ShowAlphaChannel)
            ok = dlg.exec_()
            if ok:
                self.setQColor(dlg.selectedColor())

    def qColor(self) -> QtGui.QColor:
        return self._qcolor

    def setQColor(self, color: QtGui.QColor) -> None:
        old_color = self._qcolor
        self._qcolor = color
        if self._qcolor.getRgb() != old_color.getRgb():
            self.colorChanged.emit()


class QColorLineEdit(QtW.QLineEdit):
    colorChanged = Signal()

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self._qcolor = QtGui.QColor(255, 255, 255, 255)
        self.editingFinished.connect(self._emit_color_changed)

    def qColor(self) -> QtGui.QColor:
        """Get color as QtGui.QColor object"""
        return self._qcolor

    def setQColor(self, color: QtGui.QColor):
        self._qcolor = color
        text = rgba_to_str(color.getRgb())
        self.setText(text)

    def _emit_color_changed(self):
        text = self.text()
        try:
            rgba = normalize_color(text)
        except ValueError:
            self.setQColor(self._qcolor)
        else:
            if self._qcolor.getRgb() != rgba:
                self._qcolor = QtGui.QColor(*rgba)
                self.colorChanged.emit()


class QColorEdit(QtW.QWidget):
    colorChanged = Signal(tuple)

    def __init__(self, parent=None):
        super().__init__(parent)
        _layout = QtW.QHBoxLayout()
        self._color_swatch = QColorSwatch(self)
        self._line_edit = QColorLineEdit(self)
        _layout.addWidget(self._color_swatch)
        _layout.addWidget(self._line_edit)
        self.setLayout(_layout)

        self._color_swatch.colorChanged.connect(self._on_swatch_changed)
        self._line_edit.colorChanged.connect(self._on_line_edit_edited)

    def qColor(self):
        """Return the current color."""
        return self._color_swatch._qcolor

    def color(self) -> tuple[int, int, int, int]:
        return self._color_swatch.rgba

    def setColor(self, color: QtGui.QColor | str | Iterable[int]):
        """Set value as the current color."""
        if not isinstance(color, QtGui.QColor):
            color = QtGui.QColor(*normalize_color(color))
        self._line_edit.setQColor(color)
        self._color_swatch.setQColor(color)

    def _on_line_edit_edited(self):
        self._line_edit.blockSignals(True)
        qcolor = self._line_edit.qColor()
        self._color_swatch.setQColor(qcolor)
        self._line_edit.blockSignals(False)
        self.colorChanged.emit(qcolor.getRgb())

    def _on_swatch_changed(self):
        self._color_swatch.blockSignals(True)
        qcolor = self._color_swatch.qColor()
        self._line_edit.setQColor(qcolor)
        self._color_swatch.blockSignals(False)
        self.colorChanged.emit(qcolor.getRgb())


class _ColorEdit(QBaseValueWidget):
    _qwidget: QColorEdit

    def __init__(self, **kwargs):
        super().__init__(QColorEdit, "color", "setColor", "colorChanged", **kwargs)


class ColorEdit(ValueWidget):
    """
    A widget for editing colors.

    Parameters
    ----------
    value : tuple of int or str
        RGBA color, color code or standard color name.
    """

    def __init__(self, **kwargs):
        app = use_app()
        assert app.native
        kwargs["widget_type"] = _ColorEdit
        super().__init__(**kwargs)



from __future__ import annotations
from typing import TYPE_CHECKING
from enum import Enum
from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt
from superqt.utils import WorkerBase, FunctionWorker, GeneratorWorker
from .._progress import QCircularProgressBar
from .._titlebar import QTitleBar

if TYPE_CHECKING:
    from ._tabwidget import QTabbedTableStack


class Anchor(Enum):
    """Anchor position"""

    top_left = "top_left"
    top_right = "top_right"
    bottom_left = "bottom_left"
    bottom_right = "bottom_right"


class _QOverlayBase(QtW.QDialog):
    def __init__(self, parent: QTabbedTableStack):
        """The overlay widget appears at the fixed position."""
        super().__init__(parent, Qt.WindowType.SubWindow)
        self._widget = None

        titlebar = QTitleBar("", self)
        titlebar.closeSignal.connect(self.hide)
        self._title_bar = titlebar
        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(2, 2, 2, 2)
        _layout.setSpacing(0)
        _layout.addWidget(titlebar)

        self.setLayout(_layout)

        parent.resizedSignal.connect(self.alignToParent)
        self.setAnchor(Anchor.bottom_right)
        self.setVisible(False)

    def addWidget(self, widget: QtW.QWidget):
        """Set the central widget."""
        if self._widget is not None:
            self.removeWidget()
        self.layout().addWidget(widget)
        self.resize(widget.sizeHint() + self._title_bar.sizeHint())
        self._widget = widget
        self.alignToParent()

    def removeWidget(self):
        """Remove the central widget."""
        self._widget.setParent(None)
        self._widget = None
        self.resize(QtCore.QSize(0, 0))

    def widget(self) -> QtW.QWidget:
        """The central widget."""
        return self._widget

    def anchor(self) -> Anchor:
        """Anchor position."""
        return self._anchor

    def setAnchor(self, anc: Anchor | str) -> None:
        """Set anchor position of the overlay widget."""
        self._anchor = Anchor(anc)
        return self.alignToParent()

    def title(self) -> str:
        """Title of the overlay widget."""
        return self._title_bar.title()

    def setTitle(self, title: str) -> None:
        """Set the title of the overlay widget."""
        return self._title_bar.setTitle(title)

    def show(self):
        """Show the overlay widget with animation."""
        super().show()
        self.alignToParent()
        return None

    def hide(self) -> None:
        """Hide the overlay widget with animation."""
        self.parentWidget().parent().setCellFocus()
        return super().hide()

    def alignToParent(self):
        """Position widget at the bottom right edge of the parent."""
        if not self.isVisible():
            return
        qtable = self.parentWidget()
        # if not qtable or qtable.isEmpty():
        if not qtable:
            return
        if self._anchor == Anchor.bottom_left:
            self.alignBottomLeft()
        elif self._anchor == Anchor.bottom_right:
            self.alignBottomRight()
        elif self._anchor == Anchor.top_left:
            self.alignTopLeft()
        elif self._anchor == Anchor.top_right:
            self.alignTopRight()
        else:
            raise RuntimeError

    def viewRect(self) -> QtCore.QRect:
        """Return the parent table rect."""
        parent = self.parentWidget()
        if parent.isEmpty():
            rect = parent.rect()
        else:
            qtable = parent.tableAtIndex(parent.currentIndex())
            wdt = qtable.widget(0)
            if wdt is None:
                rect = qtable.rect()
            else:
                rect = qtable.widget(0).rect()
        return rect

    def alignTopLeft(self, offset=(8, 8)):
        pos = self.viewRect().topLeft()
        pos.setX(pos.x() + offset[0])
        pos.setY(pos.y() + offset[1])
        self.move(pos)

    def alignTopRight(self, offset=(26, 8)):
        pos = self.viewRect().topRight()
        pos.setX(pos.x() - self.rect().width() - offset[0])
        pos.setY(pos.y() + offset[1])
        self.move(pos)

    def alignBottomLeft(self, offset=(8, 8)):
        pos = self.viewRect().bottomLeft()
        pos.setX(pos.x() + offset[0])
        pos.setY(pos.y() - self.rect().height() - offset[1])
        self.move(pos)

    def alignBottomRight(self, offset=(26, 8)):
        pos = self.viewRect().bottomRight()
        pos.setX(pos.x() - self.rect().width() - offset[0])
        pos.setY(pos.y() - self.rect().height() - offset[1])
        self.move(pos)

    # fmt: off
    if TYPE_CHECKING:
        def parentWidget(self) -> QTabbedTableStack: ...
    # fmt: on


class QOverlayWidget(_QOverlayBase):
    """The overlay widget appears at the fixed position."""

    def __init__(self, parent: QTabbedTableStack, duration: int = 50):
        """
        The overlay widget appears at the fixed position.

        Parameters
        ----------
        parent : QTabbedTableStack
            Parent table stack
        duration : int, default is 50
            Animation duration in msec.
        """
        super().__init__(parent)

        effect = QtW.QGraphicsOpacityEffect(self)
        effect.setOpacity(0.9)
        self.setGraphicsEffect(effect)
        self._effect = effect
        self.opacity_anim = QtCore.QPropertyAnimation(self._effect, b"opacity", self)
        self._duration = duration
        self._timer: QtCore.QTimer | None = None

    def hideLater(self, sec: float = 5):
        """Hide overlay widget after a delay."""
        self._timer = QtCore.QTimer(self)
        self._timer.setInterval(int(sec * 1000))
        self._timer.setSingleShot(True)
        self._timer.timeout.connect(self._hide)
        self._timer.start()
        return None

    def _hide(self):
        if self.isVisible():
            self.setVisible(False)
            self._timer = None
        return None

    def show(self):
        """Show the overlay widget with animation."""
        super().show()
        self.opacity_anim.setDuration(self._duration)
        self.opacity_anim.setStartValue(0)
        self.opacity_anim.setEndValue(0.9)
        self.opacity_anim.start()
        return None

    def hide(self) -> None:
        """Hide the overlay widget with animation."""
        self.parentWidget().parent().setCellFocus()
        self.opacity_anim.setDuration(self._duration)
        self.opacity_anim.setStartValue(0.9)
        self.opacity_anim.setEndValue(0)
        self.opacity_anim.start()

        @self.opacity_anim.finished.connect
        def _on_vanished():
            if self.isVisible():
                self.setVisible(False)
            self.opacity_anim.finished.disconnect()

        return None

    def enterEvent(self, a0: QtCore.QEvent) -> None:
        if self._timer is not None:
            self._timer.stop()
        return super().enterEvent(a0)

    def leaveEvent(self, a0: QtCore.QEvent) -> None:
        if self._timer is not None:
            self._timer.start()
        return super().leaveEvent(a0)


class QInfoStack(_QOverlayBase):
    def __init__(self, parent: QTabbedTableStack):
        super().__init__(parent)
        self._list_widget = QtW.QListWidget()
        self.addWidget(self._list_widget)
        self.setAnchor(Anchor.bottom_left)

    def addWorker(self, worker: WorkerBase, desc: str, total: int = 0):
        pbar = QCircularProgressBar()
        pbar.setButtonState("square")
        if isinstance(worker, FunctionWorker):
            pbar.setValue(-1)

        elif isinstance(worker, GeneratorWorker):
            _nyield = 0

            @worker.yielded.connect
            def _increment(*_):
                nonlocal _nyield
                _nyield += 1
                if _nyield > total:
                    value = -1
                else:
                    value = _nyield / total * 100
                return pbar.setValue(value)

        else:
            raise TypeError(f"Unsupported worker type: {type(worker)}")

        @pbar.abortRequested.connect
        def _aborting():
            if not worker.abort_requested:
                pbar.infiniteRequested.emit(True)
                worker.quit()

        item = QtW.QListWidgetItem()
        labeled_pbar = labeled_progressbar(desc, pbar)
        worker.started.connect(lambda: self._on_worker_started(item, labeled_pbar))
        worker.finished.connect(lambda: self._on_worker_finish(item))

    def _on_worker_started(self, item: QtW.QListWidgetItem, widget: QtW.QWidget):
        lw = self._list_widget
        lw.addItem(item)
        lw.setIndexWidget(lw.model().index(lw.count() - 1, 0), widget)
        self.adjustHeight()
        self.show()

    def _on_worker_finish(self, item: QtW.QListWidgetItem):
        lw = self._list_widget
        lw.takeItem(lw.row(item))
        if lw.count() == 0:
            self.hide()
        else:
            self.adjustHeight()

    def adjustHeight(self):
        height = 20 * min(3, self._list_widget.count()) + 6
        self._list_widget.setFixedHeight(height)
        self.setFixedHeight(height + self._title_bar.sizeHint().height())
        self.alignToParent()


def labeled_progressbar(label: str, pbar: QCircularProgressBar):
    w = QtW.QWidget()
    _layout = QtW.QHBoxLayout(w)
    _layout.setContentsMargins(0, 0, 0, 0)
    w.setLayout(_layout)
    label_widget = QtW.QLabel(label)
    _layout.addWidget(label_widget, alignment=Qt.AlignmentFlag.AlignLeft)
    pbar.setRadius(8)
    pbar.setBarWidth(3)
    pbar.setFixedSize(20, 20)
    _layout.addWidget(pbar, alignment=Qt.AlignmentFlag.AlignRight)
    return w



from __future__ import annotations

from qtpy import QtWidgets as QtW, QtGui, QtCore


class QTitleLabel(QtW.QLabel):
    """Label used for titles in the preference dialog."""

    def __init__(self, text: str, size: int) -> None:
        super().__init__()
        self.setText(text)
        self.setFont(QtGui.QFont("Arial", size))

    def paintEvent(self, a0: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self)
        color = self.palette().color(QtGui.QPalette.ColorRole.Foreground)
        painter.setPen(QtGui.QPen(color, 1))
        bottom_left = self.rect().bottomLeft()
        bottom_right = QtCore.QPoint(bottom_left.x() + 300, bottom_left.y())
        painter.drawLine(bottom_left, bottom_right)
        return super().paintEvent(a0)



from ._mainwidgets import QMainWindow, QMainWidget, _QtMainWidgetBase

# activate command palette and key combo
from . import _command_palette, _keycombo

del _command_palette, _keycombo

__all__ = ["QMainWindow", "QMainWidget", "_QtMainWidgetBase"]



from __future__ import annotations

from typing import Hashable, TYPE_CHECKING, NamedTuple
from . import _dialogs


if TYPE_CHECKING:
    from tabulous.widgets import TableBase, TableViewerBase


def plot(viewer: TableViewerBase):
    """Run plt.plot"""
    return _plot_xy(viewer, _dialogs.plot)


def bar(viewer: TableViewerBase):
    """Run plt.bar"""
    return _plot_xy(viewer, _dialogs.bar)


def scatter(viewer: TableViewerBase):
    """Run plt.scatter"""
    return _plot_xy(viewer, _dialogs.scatter)


def fill_between(viewer: TableViewerBase):
    """Run plt.fill_between"""
    table = viewer.current_table
    if table is None:
        return

    _dialogs.fill_between(
        ax={"bind": table.plt.gca()},
        table={"bind": table},
        alpha={"min": 0, "max": 1, "step": 0.05},
        parent=viewer._qwidget,
    )


def fill_betweenx(viewer: TableViewerBase):
    """Run plt.fill_betweenx"""
    table = viewer.current_table
    if table is None:
        return

    _dialogs.fill_betweenx(
        ax={"bind": table.plt.gca()},
        table={"bind": table},
        alpha={"min": 0, "max": 1, "step": 0.05},
        parent=viewer._qwidget,
    )


def errorbar(viewer: TableViewerBase):
    """Run plt.errorbar"""
    table = viewer.current_table
    if table is None:
        return

    _dialogs.errorbar(
        ax={"bind": table.plt.gca()},
        table={"bind": table},
        alpha={"min": 0, "max": 1, "step": 0.05},
        parent=viewer._qwidget,
    )


def hist(viewer: TableViewerBase):
    """Run plt.hist"""
    table = viewer.current_table
    if table is None:
        return

    _dialogs.hist(
        ax={"bind": table.plt.gca()},
        table={"bind": table},
        alpha={"min": 0, "max": 1, "step": 0.05},
        histtype={
            "choices": ["bar", "step", "stepfilled", "barstacked"],
            "value": "bar",
        },
        parent=viewer._qwidget,
    )


def swarmplot(viewer: TableViewerBase):
    """Run sns.swarmplot"""
    return _plot_sns(viewer, _dialogs.swarmplot)


def barplot(viewer: TableViewerBase):
    """Run sns.barplot"""
    return _plot_sns(viewer, _dialogs.barplot)


def boxplot(viewer: TableViewerBase):
    """Run sns.boxplot"""
    return _plot_sns(viewer, _dialogs.boxplot)


def boxenplot(viewer: TableViewerBase):
    """Run sns.boxenplot"""
    return _plot_sns(viewer, _dialogs.boxenplot)


def violinplot(viewer: TableViewerBase):
    """Run sns.violinplot"""
    return _plot_sns(viewer, _dialogs.violinplot)


def stripplot(viewer: TableViewerBase):
    """Run sns.stripplot"""
    return _plot_sns(viewer, _dialogs.stripplot)


def new_figure(viewer: TableViewerBase):
    """New figure canvas"""
    table = viewer.current_table
    if table is None:
        return
    return table.plt.new_widget()


def _plot_xy(viewer: TableViewerBase, dialog):
    table = viewer.current_table
    if table is None:
        return

    _x = _y = None
    if table.columns.size == 1:
        _y = (slice(None), slice(0, 1))
    elif table.columns.size >= 2:
        _x = (slice(None), slice(0, 1))
        _y = (slice(None), slice(1, 2))

    dialog(
        ax={"bind": table.plt.gca()},
        x={"format": "iloc", "value": _x},
        y={"format": "iloc", "value": _y},
        table={"bind": table},
        alpha={"min": 0, "max": 1, "step": 0.05},
        parent=viewer._qwidget,
    )


def _plot_sns(viewer: TableViewerBase, dialog):
    table = viewer.current_table
    if table is None:
        return

    names, csel = _get_selected_ranges_and_column_names(table)

    # infer x, y
    if len(names) == 0:
        raise ValueError("Table must have at least one column.")
    elif len(names) == 1:
        x = {"bind": None}
        y = {"bind": None}
    else:
        for i, dtype in enumerate(table.data.dtypes):
            if dtype == "categorical":
                x = {"choices": names, "value": names[i], "nullable": True}
                j = 0 if i > 0 else 1
                y = {"choices": names, "value": names[j], "nullable": True}
                break
        else:
            x = {"choices": names, "value": None, "nullable": True}
            y = {"choices": names, "value": None, "nullable": True}

    if dialog(
        ax={"bind": table.plt.gca()},
        x=x,
        y=y,
        table={"bind": table},
        csel={"bind": csel},
        hue={"choices": names, "nullable": True},
        parent=viewer._qwidget,
    ):
        table.plt.draw()


class PlotInfo(NamedTuple):
    names: list[Hashable]
    columns: slice


def _get_selected_ranges_and_column_names(
    table: TableBase,
) -> PlotInfo:
    sels = table.selections
    if len(sels) == 1 and _selection_area(sels[0]) == 1:
        nrow = len(table.index)
        infos = PlotInfo(
            names=list(table.columns),
            columns=slice(0, nrow),
        )
    else:
        names = []
        sl = sels[0][0]
        columns = table.columns
        for sel in sels:
            if sel[0] == sl:
                csel = sel[1]
                for i in range(csel.start, csel.stop):
                    names.append(columns[i])
            else:
                raise ValueError("Selections must be in the same rows")
        infos = PlotInfo(names=names, columns=sl)

    return infos


def _selection_area(sel: tuple[slice, slice]) -> int:
    return (sel[0].stop - sel[0].start) * (sel[1].stop - sel[1].start)


del NamedTuple



from __future__ import annotations
from typing import Any
import re
import requests
from superqt.utils import thread_worker
import warnings


def get_latest_version() -> str:
    url = "https://pypi.org/pypi/tabulous/json"
    resp = requests.get(url)
    if resp.status_code != 200:
        raise RuntimeError(f"Failed to get latest version: {resp.status_code}")
    js = resp.json()
    release: dict[str, Any] = js["releases"]
    all_releases = list(release.keys())
    for r in reversed(all_releases):
        if re.match(r".*[a-zA-Z].*", r):
            continue
        return r
    return "0.0.0"


def get_current_version() -> str:
    from tabulous import __version__

    return __version__


@thread_worker
def _get_latest_version() -> str | None:
    try:
        version = get_latest_version()
        if get_current_version() >= version:
            version = None
    except Exception as e:
        warnings.warn(f"Failed to fetch latest version {type(e).__name__}: {e}")
        version = None
    return version


def get_worker():
    return _get_latest_version()



from __future__ import annotations

from typing import Callable
from pathlib import Path
from qtpy import QtGui, QtCore, QtSvg
from qtpy.QtCore import Qt
from functools import lru_cache

# These classes are mostly copied from napari/_qt/qt_resources/_svg.py
# See https://github.com/napari/napari/blob/main/napari/_qt/qt_resources/_svg.py


class SVGBufferIconEngine(QtGui.QIconEngine):
    def __init__(self, xml: str | bytes) -> None:
        if isinstance(xml, str):
            xml = xml.encode("utf-8")
        self.data = QtCore.QByteArray(xml)
        super().__init__()

    def paint(self, painter: QtGui.QPainter, rect, mode, state):
        """Paint the icon int ``rect`` using ``painter``."""
        renderer = QtSvg.QSvgRenderer(self.data)
        renderer.setAspectRatioMode(Qt.AspectRatioMode.KeepAspectRatio)
        renderer.render(painter, QtCore.QRectF(rect))

    def clone(self):
        """Required to subclass abstract QIconEngine."""
        return SVGBufferIconEngine(self.data)

    def pixmap(self, size, mode, state):
        """Return the icon as a pixmap with requested size, mode, and state."""
        img = QtGui.QImage(size, QtGui.QImage.Format.Format_ARGB32)
        img.fill(Qt.GlobalColor.transparent)
        pixmap = QtGui.QPixmap.fromImage(img, Qt.ImageConversionFlag.NoFormatConversion)
        painter = QtGui.QPainter(pixmap)
        self.paint(painter, QtCore.QRect(QtCore.QPoint(0, 0), size), mode, state)
        return pixmap


class QColoredSVGIcon(QtGui.QIcon):
    _COLOR_ARG = "#000000"

    def __init__(
        self,
        xml: str,
        color: str = "#000000",
        converter: Callable[[QtGui.QColor], QtGui.QColor] = lambda x: x,
    ) -> None:
        self._color = QtGui.QColor(color)
        col = converter(self._color)
        self._xml = xml
        colorized = xml.replace(self._COLOR_ARG, col.name())
        super().__init__(SVGBufferIconEngine(colorized))

    @lru_cache
    def colored(
        self,
        color: str = "#000000",
    ) -> QColoredSVGIcon:
        return QColoredSVGIcon(self._xml, color)

    @classmethod
    def fromfile(cls: type[QColoredSVGIcon], path: str | Path, color="#000000"):
        with open(path) as f:
            xml = f.read()
        return cls(xml, color=color)

    def color(self) -> QtGui.QColor:
        """Color of the icon"""
        return self._color

    def with_converted(
        self, converter: Callable[[QtGui.QColor], QtGui.QColor]
    ) -> QColoredSVGIcon:
        return QColoredSVGIcon(self._xml, color=self._color.name(), converter=converter)



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.backend_bases import MouseEvent, MouseButton

from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Signal

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from matplotlib.axes import Axes
    from matplotlib.legend import Legend


class InteractiveFigureCanvas(FigureCanvas):
    """A figure canvas implemented with mouse callbacks."""

    figure: Figure
    deleteRequested = Signal()
    itemPicked = Signal(object)
    clicked = Signal(object)
    doubleClicked = Signal()  # emitted *before* itemPicked event

    def __init__(self, fig):
        super().__init__(fig)
        self.pressed = None
        self.lastx_pressed = None
        self.lasty_pressed = None
        self.lastx = None
        self.lasty = None
        self.last_axis: Axes | None = None
        self._interactive = True
        fig.canvas.mpl_connect("pick_event", self._emit_pick_event)

    def _emit_pick_event(self, event):
        ax: Axes = event.mouseevent.inaxes
        if ax:
            artist = event.artist
            if ax.containers:
                for container in ax.containers:
                    # if an artist is in a container, emit the container instead
                    if artist in container:
                        self.itemPicked.emit(container)
                        break
            else:
                self.itemPicked.emit(artist)

    def wheelEvent(self, event):
        """
        Resize figure by changing axes xlim and ylim. If there are subplots, only the subplot
        in which cursor exists will be resized.
        """
        ax = self.last_axis
        if not self._interactive or not ax:
            return
        delta = event.angleDelta().y() / 120
        event = self.get_mouse_event(event)
        factor = 0.75**delta

        _zoom_x_wheel(ax, factor)
        _zoom_y_wheel(ax, factor)
        self.figure.canvas.draw()
        return None

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        """Record the starting coordinates of mouse drag."""
        mouse_event = self.get_mouse_event(event)
        self.lastx_pressed = self.lastx = mouse_event.xdata
        self.lasty_pressed = self.lasty = mouse_event.ydata
        if mouse_event.inaxes:
            self.pressed = mouse_event.button
            self.last_axis = mouse_event.inaxes
        # self.clicked.emit(mouse_event)
        return None

    def mouseMoveEvent(self, event):
        """
        Translate axes focus while dragging. If there are subplots, only the subplot in which
        cursor exists will be translated.
        """
        ax = self.last_axis
        if (
            self.pressed not in (MouseButton.LEFT, MouseButton.RIGHT)
            or self.lastx_pressed is None
            or not self._interactive
            or not ax
        ):
            return

        event = self.get_mouse_event(event)
        x, y = event.xdata, event.ydata

        if x is None or y is None:
            return None

        if self.pressed == MouseButton.LEFT:
            _translate_x(ax, self.lastx_pressed, x)
            _translate_y(ax, self.lasty_pressed, y)
        elif self.pressed == MouseButton.RIGHT:
            _zoom_x(ax, self.lastx, x)
            _zoom_y(ax, self.lasty, y)

        self.lastx, self.lasty = x, y
        self.figure.canvas.draw()
        return None

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        """Stop dragging state."""
        pos = event.pos()
        mouse_event = self.get_mouse_event(event)

        if self.lastx == mouse_event.xdata and self.lasty == mouse_event.ydata:
            # clicked, not dragged
            if self.pressed == MouseButton.LEFT:
                self.clicked.emit(mouse_event)

            elif self.pressed == MouseButton.RIGHT:
                menu = self._make_context_menu(mouse_event)
                menu.exec_(self.mapToGlobal(pos))

        self.pressed = None

        return None

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        """Adjust layout upon dougle click."""
        if not self._interactive:
            return

        self.doubleClicked.emit()
        x, y = self.mouseEventCoords(event.pos())
        button = self.buttond.get(event.button())

        if button is not None:
            # native button press event to pick artists
            self.button_press_event(x, y, button, guiEvent=event)

        self.figure.tight_layout()
        self.figure.canvas.draw()
        return None

    def resizeEvent(self, event):
        """Adjust layout upon canvas resized."""
        super().resizeEvent(event)
        self.figure.tight_layout()
        self.figure.canvas.draw()
        return None

    def get_mouse_event(self, event, name="") -> MouseEvent:
        x, y = self.mouseEventCoords(event)
        if hasattr(event, "button"):
            button = self.buttond.get(event.button())
        else:
            button = None
        mouse_event = MouseEvent(name, self, x, y, button=button, guiEvent=event)
        return mouse_event

    def _make_context_menu(self, event: MouseEvent) -> QtW.QMenu:
        """Make a QMenu object with default actions."""
        menu = QtW.QMenu(self)
        menu.addAction("Copy ...", self._copy_canvas)
        menu.addAction("Save As...", self._save_canvas_dialog)
        menu.addAction("Clear figure", self._clear_figure)
        menu.addAction("Legend", self._toggle_legend)
        menu.addSeparator()
        return menu

    def _save_canvas_dialog(self, format="PNG"):
        """Open a file dialog and save the current canvas state."""
        dialog = QtW.QFileDialog(self, "Save Image")
        dialog.setAcceptMode(QtW.QFileDialog.AcceptMode.AcceptSave)
        dialog.setDefaultSuffix(format.lower())
        dialog.setNameFilter(f"{format} file (*.{format.lower()})")
        if dialog.exec_():
            filename = dialog.selectedFiles()[0]
            self.figure.savefig(filename)

    def _clear_figure(self):
        for ax in self.figure.axes:
            ax.cla()
        self.figure.canvas.draw()

    def _toggle_legend(self):
        legend: Legend = self.last_axis.get_legend()
        if legend is None:
            self.last_axis.legend()
        else:
            legend.set_visible(not legend.get_visible())
        self.figure.canvas.draw()

    def _asarray(self) -> np.ndarray:
        """Convert current canvas state into RGBA numpy array."""
        return np.asarray(self.renderer.buffer_rgba(), dtype=np.uint8)

    def _copy_canvas(self):
        """Copy current canvas state into clipboard."""
        arr = self._asarray()
        clipboard = QtW.QApplication.clipboard()
        h, w, _ = arr.shape
        image = QtGui.QImage(arr, w, h, QtGui.QImage.Format.Format_RGBA8888)
        clipboard.setImage(image)


def _translate_x(ax: Axes, xstart: float, xstop: float):
    xscale = ax.get_xscale()
    x0, x1 = ax.get_xlim()
    if xscale == "linear":
        dx = xstop - xstart
        ax.set_xlim([x0 - dx, x1 - dx])
    elif xscale == "log":
        if xstart <= 0 or xstop <= 0:
            ax.autoscale(axis="x")
        else:
            ratio = xstart / xstop
            ax.set_xlim([x0 * ratio, x1 * ratio])


def _translate_y(ax: Axes, ystart: float, ystop: float):
    yscale = ax.get_yscale()
    y0, y1 = ax.get_ylim()
    if yscale == "linear":
        dy = ystop - ystart
        ax.set_ylim([y0 - dy, y1 - dy])
    elif yscale == "log":
        if ystart <= 0 or ystop <= 0:
            ax.autoscale(axis="y")
        else:
            ratio = ystart / ystop
            ax.set_ylim([y0 * ratio, y1 * ratio])


def _zoom_x(ax: Axes, xstart: float, xstop: float):
    xscale = ax.get_xscale()
    x0, x1 = ax.get_xlim()
    if xscale == "linear":
        _u = x1 + x0
        _v = x1 - x0
        dx = xstop - xstart
        ax.set_xlim([_u / 2 - _v / 2 + dx, _u / 2 + _v / 2 - dx])
    elif xscale == "log":
        if xstart <= 0 or xstop <= 0:
            ax.autoscale(axis="x")
        ratio = xstop / xstart
        ax.set_xlim([x0 * ratio, x1 / ratio])


def _zoom_y(ax: Axes, ystart: float, ystop: float):
    yscale = ax.get_yscale()
    y0, y1 = ax.get_ylim()
    if yscale == "linear":
        _u = y1 + y0
        _v = y1 - y0
        dy = ystop - ystart
        ax.set_ylim([_u / 2 - _v / 2 + dy, _u / 2 + _v / 2 - dy])
    elif yscale == "log":
        if ystart <= 0 or ystop <= 0:
            ax.autoscale(axis="y")
        else:
            ratio = ystop / ystart
            ax.set_ylim([y0 * ratio, y1 / ratio])


def _zoom_x_wheel(ax: Axes, factor: float):
    xscale = ax.get_xscale()
    x0, x1 = ax.get_xlim()
    if xscale == "linear":
        _u = x1 + x0
        _v = x1 - x0
        ax.set_xlim([_u / 2 - _v / 2 * factor, _u / 2 + _v / 2 * factor])
    elif xscale == "log":
        if x0 <= 0 or x1 <= 0:
            ax.autoscale(axis="x")
        else:
            xc = (x0 * x1) ** 0.5
            x0_t = (x0 / xc) ** factor
            x1_t = (x1 / xc) ** factor
            ax.set_xlim([x0_t * xc, x1_t * xc])


def _zoom_y_wheel(ax: Axes, factor: float):
    yscale = ax.get_yscale()
    y0, y1 = ax.get_ylim()
    if yscale == "linear":
        _u = y1 + y0
        _v = y1 - y0
        ax.set_ylim([_u / 2 - _v / 2 * factor, _u / 2 + _v / 2 * factor])
    elif yscale == "log":
        if y0 <= 0 or y1 <= 0:
            ax.autoscale(axis="y")
        else:
            yc = (y0 * y1) ** 0.5
            y0_t = (y0 / yc) ** factor
            y1_t = (y1 / yc) ** factor
            ax.set_ylim([y0_t * yc, y1_t * yc])



from __future__ import annotations
from typing import TYPE_CHECKING, cast
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt, Signal

if TYPE_CHECKING:
    from ._table_base import QBaseTable
    from ._enhanced_table import _QTableViewEnhanced
    from ..._table_stack import QTabbedTableStack

# Wrapper widgets that can be used to wrap a QTableView


class QTableGroup(QtW.QSplitter):
    """Split view of two tables."""

    focusChanged = Signal(int)

    def __init__(self, tables: list[QBaseTable], orientation=Qt.Orientation.Horizontal):
        super().__init__(orientation)
        self.setChildrenCollapsible(False)

        self._tables: list[QBaseTable] = []
        for i, table in enumerate(tables):
            view_copy = table._central_widget.copy(link=True)
            view_copy.focusedSignal.connect(
                lambda: self.focusChanged.emit(self.focusedIndex())
            )
            self.addWidget(view_copy)
            self.setStretchFactor(i, 1)
            self._tables.append(table)

        self.focusChanged.connect(self._on_focus_changed)

    def copy(self) -> QTableGroup:
        """Make a copy of this widget."""
        copy = self.__class__(self.tables, self.orientation())
        return copy

    @property
    def tables(self) -> list[QBaseTable]:
        """List of tables."""
        return self._tables.copy()

    def pop(self, index: int = -1) -> QBaseTable:
        """Pop a table from the group."""
        if index < 0:
            index += self.count()
        out = self._tables.pop(index)
        table = self.widget(index)  # this is a copy
        table._selection_model.moving.disconnect(table._on_moving)
        table._selection_model.moved.disconnect(table._on_moved)
        table.deleteLater()
        table.setParent(None)
        return out

    def tableIndex(self, table: QBaseTable) -> int:
        """Return the index of a table in the group."""
        model = table.model()
        for t in self.tables:
            if t.model() is model:
                return t
        raise ValueError("Table not found.")

    def focusedIndex(self) -> int:
        """Return the index of the currently focused table."""
        for i in range(self.count()):
            if self.tableHasFocus(i):
                return i
        return -1

    def setFocusedIndex(self, index: int) -> None:
        """Set the focused widget to the table at index."""
        for i in range(self.count()):
            if i != index:
                self.widget(i).setBackgroundRole(QtGui.QPalette.ColorRole.Dark)
            else:
                self.widget(i).setBackgroundRole(QtGui.QPalette.ColorRole.Light)
        return self.widget(index).setFocus()

    def focusedTable(self) -> QBaseTable | None:
        """Return the currently focused table widget."""
        i = self.focusedIndex()
        if i < 0:
            return None
        return self.tables[i]

    def tableHasFocus(self, index: int) -> bool:
        """True if the table at index has focus."""
        return self.widget(index).hasFocus()

    def __eq__(self, other: QTableGroup) -> bool:
        if not isinstance(other, QTableGroup):
            return False
        # NOTE: This should be safe. Table groups derived from the same ancestor
        # will always have exclusively the same set of tables.
        return self._tables[0].model() is other._tables[0].model()

    def _on_focus_changed(self, idx: int):
        stack = self.tableStack()
        try:
            idx_dst = stack._group_index_to_tab_index(self, idx)
        except ValueError:
            return
        dst = cast(QTableGroup, stack.widget(idx_dst))
        self.blockSignals(True)
        dst.blockSignals(True)
        stack.setCurrentIndex(idx_dst)
        dst.setFocusedIndex(idx)
        dst.blockSignals(False)
        self.blockSignals(False)

    def tableStack(self) -> QTabbedTableStack:
        return self.parent().parent()

    def deleteLater(self) -> None:
        for i in range(self.count()):
            qtableview = self.widget(i)
            qtableview._selection_model.moving.disconnect(qtableview._on_moving)
            qtableview._selection_model.moved.disconnect(qtableview._on_moved)

        return super().deleteLater()

    if TYPE_CHECKING:

        def widget(self, index: int) -> _QTableViewEnhanced:
            ...



from __future__ import annotations

from functools import lru_cache
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg

# See https://stackoverflow.com/questions/32035251/displaying-latex-in-pyqt-pyside-qtablewidget
@lru_cache(maxsize=20)
def latex_to_pixmap(latex: str, fs: int, color: str = "#000000") -> QtGui.QPixmap:
    fig = Figure()
    fig.patch.set_facecolor("none")
    fig.set_canvas(FigureCanvasAgg(fig))
    renderer = fig.canvas.get_renderer()

    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis("off")
    ax.patch.set_facecolor("none")
    t = ax.text(0, 0, latex, ha="left", va="bottom", fontsize=fs, color=color)

    fwidth, fheight = fig.get_size_inches()
    fig_bbox = fig.get_window_extent(renderer)

    text_bbox = t.get_window_extent(renderer)

    tight_fwidth = text_bbox.width * fwidth / fig_bbox.width
    tight_fheight = text_bbox.height * fheight / fig_bbox.height

    fig.set_size_inches(tight_fwidth, tight_fheight)

    buf, size = fig.canvas.print_to_buffer()
    qimage = QtGui.QImage.rgbSwapped(
        QtGui.QImage(buf, size[0], size[1], QtGui.QImage.Format.Format_ARGB32)
    )
    qpixmap = QtGui.QPixmap(qimage)

    return qpixmap


class QLatexLabel(QtW.QLabel):
    def __init__(self, latex: str, parent=None):
        super().__init__(parent)
        self._color = QtGui.QColor("#000000")
        self._latex = latex
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setWordWrap(True)
        self.setLatex(latex)

    def latex(self) -> str:
        return self._latex

    def setLatex(self, latex: str):
        self.setPixmap(
            latex_to_pixmap(latex, self.font().pointSize(), self._color.name())
        )
        self._latex = latex
        self.adjustSize()

    def textColor(self) -> QtGui.QColor:
        return self._color

    def setTextColor(self, color) -> None:
        self._color = QtGui.QColor(color)
        self.setLatex(self._latex)



from __future__ import annotations

from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt

from ._theme import QThemePanel
from ._table_config import QTableConfigPanel
from ._general import QGeneralPanel
from ._shared import QTitleLabel

if TYPE_CHECKING:
    from tabulous._qt._mainwindow import _QtMainWidgetBase


class QPreferenceDialog(QtW.QDialog):
    def __init__(self, viewer: _QtMainWidgetBase) -> None:
        super().__init__(viewer)
        self._viewer = viewer
        self.setWindowTitle("Preferences")
        self.resize(600, 400)
        layout = QtW.QHBoxLayout()
        self.setLayout(layout)

        self._list = QtW.QListWidget(self)
        self._list.setFixedWidth(150)
        self._list.setFont(QtGui.QFont("Arial", 13))
        self._stack = QtW.QStackedWidget(self)

        self._list.itemClicked.connect(
            lambda: self._stack.setCurrentIndex(self._list.currentRow())
        )

        layout.addWidget(self._list)
        layout.addWidget(self._stack)

        self._setup_panels()

    def addPanel(self, name: str):
        self._list.addItem(name)
        widget = QtW.QWidget(self._stack)
        layout = QtW.QVBoxLayout()
        widget.setLayout(layout)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self._stack.addWidget(widget)
        return widget

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if (
            a0.key() == Qt.Key.Key_W
            and a0.modifiers() & Qt.KeyboardModifier.ControlModifier
        ):
            self.close()
        return super().keyPressEvent(a0)

    def _setup_panels(self):
        panel_general = self.addPanel("General")
        panel_general.layout().addWidget(QTitleLabel("General Configuration", 18))
        panel_general.layout().addWidget(QGeneralPanel(self))
        panel_apperance = self.addPanel("Apperance")
        panel_apperance.layout().addWidget(QTitleLabel("Theme", 18))
        panel_apperance.layout().addWidget(QThemePanel(self))
        panel_table = self.addPanel("Table")
        panel_table.layout().addWidget(QTitleLabel("Table Configuation", 18))
        panel_table.layout().addWidget(QTableConfigPanel(self))



from __future__ import annotations

from enum import Enum
import numpy as np
import pandas as pd
from psygnal import Signal
from magicgui.widgets import Container, ComboBox, PushButton
from tabulous._magicgui import SelectionWidget, find_current_table


class Formula(Enum):
    linear = "linear"
    quadratic = "quadratic"
    cubic = "cubic"


_EXPRESSION = {
    Formula.linear: "y = ax + b",
    Formula.quadratic: "y = ax<sup>2</sup> + bx + c",
    Formula.cubic: "y = ax<sup>3</sup> + bx<sup>2</sup> + cx + d",
}


class PolynomialFitWidget(Container):
    called = Signal(pd.Series)

    def __init__(self):
        self._xrange = SelectionWidget(label="X")
        self._yrange = SelectionWidget(label="Y")
        self._formula = ComboBox(
            choices=[(_EXPRESSION[k], k) for k in Formula], label="Formula"
        )
        self._btn = PushButton(text="Fit")
        self._btn.changed.connect(self.fit)

        widgets = [self._xrange, self._yrange, self._formula, self._btn]
        super().__init__(widgets=widgets)

    def _get_xy(self) -> tuple[np.ndarray, np.ndarray]:
        table = find_current_table(self)
        df = table.data
        x = self._xrange.value.operate(df)
        y = self._yrange.value.operate(df)
        if 1 in x.shape and 1 in y.shape:
            return x.values.ravel(), y.values.ravel()
        raise ValueError("X and Y must be 1D array")

    def fit(self):
        x, y = self._get_xy()
        if self._formula.value is Formula.linear:
            df = self._fit_linear(x, y)
        elif self._formula.value is Formula.quadratic:
            df = self._fit_quadratic(x, y)
        elif self._formula.value is Formula.cubic:
            df = self._fit_cubic(x, y)
        else:
            raise RuntimeError("Unknown formula")
        self.close()
        self.called.emit(df)

    def _fit_linear(self, x: np.ndarray, y: np.ndarray):
        coef, (res, *_) = np.polynomial.polynomial.polyfit(x, y, 1, full=True)
        return pd.DataFrame(
            [[coef[1]], [coef[0]], [res[0]]],
            index=["a", "b", "residual"],
            columns=["result"],
        )

    def _fit_quadratic(self, x: np.ndarray, y: np.ndarray):
        coef, (res, *_) = np.polynomial.polynomial.polyfit(x, y, 2, full=True)
        return pd.DataFrame(
            [[coef[2]], [coef[1]], [coef[0]], [res[0]]],
            index=["a", "b", "c", "residual"],
            columns=["result"],
        )

    def _fit_cubic(self, x: np.ndarray, y: np.ndarray):
        coef, (res, *_) = np.polynomial.polynomial.polyfit(x, y, 3, full=True)
        return pd.DataFrame(
            [[coef[3]], [coef[2]], [coef[1]], [coef[0]], [res[0]]],
            index=["a", "b", "c", "d", "residual"],
            columns=["result"],
        )



from ._table import TableBase, Table, SpreadSheet, GroupBy, TableDisplay
from ._mainwindow import TableViewer, TableViewerWidget, TableViewerBase
from ._magicgui import MagicTable, MagicSpreadSheet

__all__ = [
    "TableBase",
    "Table",
    "SpreadSheet",
    "GroupBy",
    "TableDisplay",
    "TableViewer",
    "TableViewerWidget",
    "TableViewerBase",
    "MagicTable",
    "MagicSpreadSheet",
]



from __future__ import annotations
from typing import TYPE_CHECKING

from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt

from tabulous._qt._titlebar import QTitleBar


class QInnerSplitter(QtW.QSplitter):
    """The inner widget for QTableSideArea."""

    def __init__(self, parent: QtW.QWidget = None) -> None:
        super().__init__(Qt.Orientation.Vertical, parent)

    def sizeHint(self):
        nwidgets = self.count()
        if self.orientation() == Qt.Orientation.Vertical:
            w, h = self.width(), 0
            for i in range(nwidgets):
                child = self.widget(i)
                h += child.minimumHeight()
                w = min(w, child.minimumWidth())
        else:
            w, h = 0, self.height()
            for i in range(nwidgets):
                child = self.widget(i)
                h = max(h, child.minimumHeight())
                w += child.minimumWidth()
        return QtCore.QSize(w, h)


class QTableSideArea(QtW.QScrollArea):
    """The side scroll area of a table."""

    def __init__(self, parent: QtW.QWidget | None = None) -> None:
        super().__init__(parent)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setWidgetResizable(True)
        widget_inside = QInnerSplitter(self)
        self.setMinimumWidth(180)
        self.setMinimumHeight(180)
        self.setWidget(widget_inside)
        self._widgets: list[QtW.QWidget] = []
        # add empty widget
        splitter = self.widget()
        splitter.addWidget(QtW.QWidget())

    # fmt: off
    if TYPE_CHECKING:
        def widget(self) -> QInnerSplitter: ...
        def parentWidget(self) -> QtW.QSplitter: ...
    # fmt: on

    def setOrientation(self, orientation: Qt.Orientation) -> None:
        self.widget().setOrientation(orientation)
        if orientation == Qt.Orientation.Vertical:
            self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        else:
            self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)

    def addWidget(self, widget: QtW.QWidget, name: str = "") -> None:
        if widget in self._widgets:
            return

        dock = QSplitterDockWidget(widget, name=name)
        splitter = self.widget()
        idx = splitter.count() - 1
        splitter.insertWidget(idx, dock)
        self._widgets.append(widget)
        splitter.setCollapsible(idx, False)

        @dock._title_bar.closeSignal.connect
        def _():
            self.removeWidget(widget)
            # NOTE: if not deleted, the widget will keep emitting signals
            widget.deleteLater()

    def removeWidget(self, widget: QtW.QWidget) -> None:
        """Remove given widget from the side area."""
        idx = -1
        for i, wdt in enumerate(self._widgets):
            if wdt is widget:
                # NOTE: To avoid the child of the dock widget get deleted,
                # it must be removed from its parent before the removal of
                # the dock widget.
                dock = wdt.parent()
                wdt.setParent(None)
                dock.setParent(None)
                idx = i
                break
        else:
            raise RuntimeError("Widget not found in the list.")

        del self._widgets[idx]

        if len(self._widgets) == 0:
            self.parentWidget().setSizes([1, 0])


class QSplitterDockWidget(QtW.QWidget):
    """The dock widget for QTableSideArea"""

    def __init__(self, widget: QtW.QWidget, name: str = "") -> None:
        super().__init__()
        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(0, 0, 0, 0)
        _layout.setSpacing(0)
        title = QTitleBar(title=name, parent=self)
        _layout.addWidget(title)
        _layout.addWidget(widget)
        self.setLayout(_layout)

        self._title_bar = title
        self._widget = widget

    @property
    def close_btn(self):
        return self._title_bar._close_button
